#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2011 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Upgrade SchoolTool to generation 37.

Move per-schoolyear levels to a single per-app container.
"""
from zope.app.generations.utility import findObjectsProviding
from zope.app.publication.zopepublication import ZopePublication
from zope.component import getUtility
from zope.component.hooks import getSite, setSite
from zope.intid.interfaces import IIntIds

from schooltool.app.interfaces import ISchoolToolApplication
from schooltool.level.level import LevelContainerContainer
from schooltool.schoolyear.interfaces import ISchoolYearContainer


LEVELS_APP_KEY = 'schooltool.level.level'


def guessMostRecentLevels(app):
    container = app.get(LEVELS_APP_KEY)
    if (container is None or
        not isinstance(container, LevelContainerContainer)):
        return None
    levels = None
    syc = ISchoolYearContainer(app, None)
    int_ids = getUtility(IIntIds)

    years = list(reversed(syc.sorted_schoolyears))
    active_year = syc.getActiveSchoolYear()
    if active_year is not None:
        years = [active_year] + years
    for year in years:
        sy_id = str(int_ids.getId(year))
        levels = container.get(sy_id, None)
        if levels is not None:
            return levels
    return None


def evolve(context):
    root = context.connection.root().get(ZopePublication.root_name, None)
    old_site = getSite()

    apps = findObjectsProviding(root, ISchoolToolApplication)
    for app in apps:
        setSite(app)
        levels = guessMostRecentLevels(app)
        if levels is not None:
            del app[LEVELS_APP_KEY]
            app[LEVELS_APP_KEY] = levels

    setSite(old_site)



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2005 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Tests for DateFormatter views.
"""
import unittest
import doctest
import datetime

from zope.publisher.browser import TestRequest

from schooltool.app.browser.testing import setUp, tearDown
from schooltool.skin.dateformatter import DateFormatterFullView
from schooltool.skin.dateformatter import DateFormatterMediumView
from schooltool.skin.dateformatter import DateFormatterShortView
from schooltool.skin.dateformatter import LocaleLookupMixin


def doctest_DateFormatterView():
    r"""Test for DateFormatterView

    Let's say we need to localize the 1 december 2006

    We create 2 clients, one english, one french:

        >>> english_request = TestRequest(environ={'HTTP_ACCEPT_LANGUAGE':'en'})
        >>> french_request = TestRequest(environ={'HTTP_ACCEPT_LANGUAGE':'fr,en'})

    We setup the date:

        >>> date = datetime.date(2006, 12, 1)
        >>> date
        datetime.date(2006, 12, 1)
        >>> isinstance(date, datetime.date)
        True

    We take the Full format of a date:

        >>> view = DateFormatterFullView(date, english_request)
        >>> view()
        u'Friday, December 1, 2006'

    and in french:

        >>> view = DateFormatterFullView(date, french_request)
        >>> view()
        u'vendredi 1 d\xe9cembre 2006'

    We take the medium format of a date

        >>> view = DateFormatterMediumView(date, english_request)
        >>> view()
        u'Dec 1, 2006'

    and in french:

        >>> view = DateFormatterMediumView(date, french_request)
        >>> view()
        u'1 d\xe9c. 06'

    We take the short format of a date

        >>> view = DateFormatterShortView(date, english_request)
        >>> view()
        '2006-12-01'

    even in french the short format is ISO format:

        >>> view = DateFormatterShortView(date, french_request)
        >>> view()
        '2006-12-01'

    """

def doctest_LocaleLookupMixin():
    r"""Tests for LocaleLookupMixin

    Let's try first to use the mixin without a view and thus without
    a request attribute. This will fail as there is no request attribute
    provided

        >>> mixin = LocaleLookupMixin()
        >>> mixin.getLocale()
        Traceback (most recent call last):
            ...
        NotImplementedError: LocaleLookupMixin need to be applied on a view

    Now let's play with a real view which will use our LocaleLookupMixin

        >>> from zope.publisher.browser import BrowserView
        >>> class FooView(BrowserView, LocaleLookupMixin):
        ...     pass

    Let's create a dummy context for the view

        >>> class FooContext(object):
        ...     pass
        >>> barObj = FooContext()

    Let's create two simple request

        >>> french_request = TestRequest(environ={'HTTP_ACCEPT_LANGUAGE':'fr,en'})

    We can now instanciate the view

        >>> v = FooView(barObj, french_request)

        >>> locale = v.getLocale()

    We have then an instance of the Locale class

        >>> locale
        <zope.i18n.locales.Locale object ...>

    Which is french as in our request

        >>> locale.getLocaleID()
        u'fr'

    Now let's see what happen is no locale is defined in the request

        >>> simple_request = TestRequest()

    We don't have any language defined:

        >>> simple_request.locale.id.language is None
        True

    Let's see how the Mixin will behave, it should return an english
    locale by default:

        >>> v = FooView(barObj, simple_request)
        >>> locale = v.getLocale()
        >>> locale
        <zope.i18n.locales.Locale object ...>
        >>> locale.getLocaleID()
        'en_US'
    """

def test_suite():
    suite = unittest.TestSuite()
    suite.addTest(doctest.DocTestSuite(
        setUp=setUp, tearDown=tearDown,
        optionflags=doctest.ELLIPSIS|doctest.REPORT_NDIFF))
    return suite

if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2011 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Schooltool page AJAX parts.
"""
from zope.component import adapts
from zope.interface import implements
from zope.location.interfaces import LocationError
from zope.publisher.interfaces import NotFound
from zope.publisher.interfaces.browser import IBrowserPublisher
from zope.traversing.api import traverseName
from zope.traversing.interfaces import ITraversable

from schooltool.skin.flourish import interfaces
from schooltool.skin.flourish import tal
from schooltool.skin.flourish.viewlet import Viewlet, ViewletManagerBase
from schooltool.skin.flourish.viewlet import ManagerViewlet


class AJAXPart(Viewlet):
    implements(interfaces.IAJAXPart, IBrowserPublisher)

    fromPublication = False

    @property
    def ignoreRequest(self):
        return not self.fromPublication

    def browserDefault(self, request):
        self.fromPublication = True
        return self, ()

    def publishTraverse(self, request, name):
        raise NotFound(self, name, request)

    def setJSONResponse(self, data):
        """
        Switch response to JSON.  Return encoded payload.
        Resets the original response. Yes, I know, this method is evil.
        Sorry.
        """
        response = self.request.response

        # Bye bye birdie.
        response.reset()

        response.setHeader('Content-Type', 'application/json')
        encoder = tal.JSONEncoder()
        json = encoder.encode(data)
        return json


class CompositeAJAXPart(ManagerViewlet, AJAXPart):
    implements(interfaces.IAJAXPart, IBrowserPublisher)

    def publishTraverse(self, request, name):
        part = self.get(name)
        if part is None:
            raise NotFound(self, name, request)
        return part


class AJAXParts(ViewletManagerBase):
    implements(interfaces.IAJAXParts)

    def render(self, *args, **kw):
        raise NotFound(self.__parent__, self.__name__, self.request)

    def publishTraverse(self, request, name):
        try:
            part = traverseName(self, name)
        except LocationError:
            raise NotFound(self, name, self.request)
        return part


class AJAXPartsTraversable(object):
    adapts(AJAXParts)
    implements(ITraversable)

    def __init__(self, parts):
        self.parts = parts

    def traverse(self, name, furtherPath):
        __traceback_info__ = (self.parts, name, furtherPath)
        try:
            return self.parts[name]
        except (KeyError, TypeError):
            raise LocationError(self.parts, name)


class SubContextParts(CompositeAJAXPart, AJAXParts):

    render = AJAXParts.render
    publishTraverse = AJAXParts.publishTraverse


class ViewContextParts(SubContextParts):

    def __init__(self, context, request, view, manager):
        SubContextParts.__init__(self, view, request, view, manager)


class ContextTraversable(object):
    adapts(SubContextParts)
    implements(ITraversable)

    def __init__(self, parts):
        self.parts = parts

    def traverse(self, name, furtherPath):
        __traceback_info__ = (self.parts, name, furtherPath)
        try:
            return self.parts[name]
        except (KeyError, TypeError):
            pass

        try:
            next = traverseName(self.parts.context, name)
        except LocationError:
            raise

        parts = SubContextParts(
            next, self.parts.request, self.parts.view, self.parts)
        parts.__name__ = name
        return parts



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2006 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
SchoolTool security policy.

"""
from persistent import Persistent
import zope.keyreference.interfaces
from zope.security.simplepolicies import ParanoidSecurityPolicy
from zope.security.proxy import removeSecurityProxy
from zope.component import queryAdapter
from zope.traversing.api import getParent
from schooltool.securitypolicy.crowds import ICrowd
from schooltool.securitypolicy.metaconfigure import getCrowdsUtility


class SchoolToolSecurityPolicy(ParanoidSecurityPolicy):
    """Crowd-based security policy."""

    def checkPermission(self, permission, obj):
        """Return True if principal has permission on object."""

        # Check the generic, interface-independent permissions.
        factories = getCrowdsUtility().getFactories(permission, None)
        if factories:
            return self.checkCrowds(factories, obj)

        # No quick method worked, look up the crowd by adaptation.
        return self.checkByAdaptation(permission, obj)

    def checkByAdaptation(self, permission, obj):
        crowd = queryAdapter(obj, ICrowd, name=permission, default=None)
        # If there is no crowd that has the given permission on this
        # object, try to look up a crowd that includes the parent.
        while crowd is None and obj is not None:
            obj = getParent(obj)
            crowd = queryAdapter(obj, ICrowd, name=permission, default=None)
        if crowd is None: # no crowds found
            raise AssertionError('no crowd found for', obj, permission)

        for participation in self.participations:
            if crowd.contains(participation.principal):
                return True
        else:
            return False

    def checkCrowds(self, factories, obj):
        """Check if an object is in any of the given crowds."""
        for participation in self.participations:
            for factory in factories:
                crowd = factory(obj)
                if crowd.contains(participation.principal):
                    return True
        return False


class CachingSecurityPolicy(ParanoidSecurityPolicy):
    """Crowd-based caching security policy."""

    def cachingKey(self, permission, obj):
        try:
            ref = zope.keyreference.interfaces.IKeyReference(obj, None)
            if ref is None:
                return None
        except zope.keyreference.interfaces.NotYet:
            return None
        return (permission, ref)

    @classmethod
    def getCache(cls, participation):
        cache = getattr(participation, '_st_perm_cache', None)
        if cache is None:
            try:
                participation._st_perm_cache = {'perm': {},
                                                'enabled': True}
            except AttributeError:
                return None
        return participation._st_perm_cache

    def checkCache(self, permission, obj):
        key = self.cachingKey(permission, obj)
        if key is None:
            return None # uncacheable
        result = None
        for participation in self.participations:
            cache = self.getCache(participation)
            if cache is not None:
                perm = cache['perm'].get(key, None)
                result = max(result, perm)
        return result

    def cache(self, participation, permission, obj, value):
        cache = self.getCache(participation)
        if cache is None or not cache['enabled']:
            return
        key = self.cachingKey(permission, obj)
        if key is None:
            return # uncacheable
        cache['perm'][key] = value

    def checkPermission(self, permission, obj):
        """Return True if principal has permission on object."""

        perm = self.checkCache(permission, obj)
        if perm is not None:
            return perm

        perm = self.checkPermissionCrowds(permission, obj)
        if perm is not None:
            return perm

        perm = self.checkByAdaptation(permission, obj)
        return perm

    def checkByAdaptation(self, permission, obj):
        crowd = queryAdapter(obj, ICrowd, name=permission, default=None)
        # If there is no crowd that has the given permission on this
        # object, try to look up a crowd that includes the parent.
        objects = [obj]
        while crowd is None and obj is not None:
            obj = getParent(obj)
            objects.append(obj)
            crowd = queryAdapter(obj, ICrowd, name=permission, default=None)
        if crowd is None: # no crowds found
            raise AssertionError('no crowd found for', obj, permission)

        for participation in self.participations:
            if crowd.contains(participation.principal):
                for o in objects:
                    self.cache(participation, permission, o, True)
                return True
            else:
                for o in objects:
                    self.cache(participation, permission, o, False)
        else:
            return False

    def checkPermissionCrowds(self, permission, obj):
        """Check object-independent crowds."""
        factories = getCrowdsUtility().getFactories(permission, None)
        if not factories:
            return None

        for participation in self.participations:
            for factory in factories:
                crowd = factory(obj)
                if crowd.contains(participation.principal):
                    self.cache(participation, permission, obj, True)
                    return True
            self.cache(participation, permission, obj, False)
        return False



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2011 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
AJAX-style tables.
"""

from zope.interface import implements
from zope.browserpage.viewpagetemplatefile import ViewPageTemplateFile
from zope.cachedescriptors.property import Lazy

import zc.resourcelibrary
from zc.table import table

from schooltool.common.inlinept import InlineViewPageTemplate
from schooltool.skin import flourish
from schooltool.table.interfaces import IFilterWidget
from schooltool.table.interfaces import IIndexedColumn
from schooltool.table.batch import TokenBatch
from schooltool.table.table import TableContent, FilterWidget
from schooltool.table.table import url_cell_formatter
from schooltool.table.table import SortUIHeaderMixin
from schooltool.table.table import HeaderFormatterMixin
from schooltool.table.table import StandaloneHeaderFormatterMixin
from schooltool.table.catalog import IndexedTableFormatter
from schooltool.table.catalog import IndexedFilterWidget
from schooltool.common import SchoolToolMessage as _


class AJAXSortHeaderMixin(SortUIHeaderMixin):

    html_id = None

    def _header_template(self, options):
        options = dict(options)
        options['containerID'] = self.html_id
        template = """
            <span class="%(css_class)s"
                  onclick="javascript: %(script_name)s(
                        '%(containerID)s', '%(columnName)s', '%(sort_on_name)s')">
                %(header)s</span>
        """
        return template % options


class AJAXFormSortFormatter(HeaderFormatterMixin,
                            AJAXSortHeaderMixin,
                            table.FormSortFormatter):
    script_name = 'ST.table.on_form_sort'


class AJAXStandaloneSortFormatter(StandaloneHeaderFormatterMixin,
                                  AJAXSortHeaderMixin,
                                  table.StandaloneSortFormatter):
    script_name = 'ST.table.on_standalone_sort'


class Table(flourish.ajax.CompositeAJAXPart, TableContent):

    no_default_url_cell_formatter = False

    container_wrapper = ViewPageTemplateFile('templates/f_ajax_table.pt')

    form_wrapper = InlineViewPageTemplate("""
      <form method="post" tal:attributes="action view/@@absolute_url;
                                          class view/form_class;
                                          id view/form_id">
        <tal:block replace="structure view/template" />
      </form>
    """)

    empty_message = u""
    form_class = None
    form_id = None

    table_formatter = AJAXFormSortFormatter

    inside_form = False # don't surround with <form> tag if inside_form

    @Lazy
    def html_id(self):
        return flourish.page.generic_viewlet_html_id(self, self.prefix)

    @Lazy
    def filter_widget(self):
        return self.get('filter')

    @Lazy
    def batch(self):
        return self.get('batch')

    def updateFormatter(self):
        if self._table_formatter is not None:
            return
        if self.no_default_url_cell_formatter:
            formatters = []
        else:
            formatters = [url_cell_formatter]
        self.setUp(formatters=formatters,
                   table_formatter=self.table_formatter,
                   batch_size=self.batch_size,
                   prefix=self.__name__,
                   css_classes={'table': 'data'})

    def update(self):
        self.updateFormatter()
        TableContent.update(self)
        flourish.ajax.CompositeAJAXPart.update(self)

    def makeFormatter(self):
        if self._table_formatter is None:
            return None
        formatter = self._table_formatter(
            self.source, self.request, self._items,
            visible_column_names=self.visible_column_names,
            columns=self._columns,
            batch_start=self.batch.start, batch_size=self.batch.size,
            sort_on=self._sort_on,
            prefix=self.prefix,
            ignore_request=self.ignoreRequest,
            group_by_column=self.group_by_column,
            )
        formatter.html_id = self.html_id
        formatter.view = self
        formatter.cssClasses.update(dict(self.css_classes))
        return formatter

    def renderTable(self):
        formatter = self.makeFormatter()
        return formatter() if formatter is not None else ''

    def render(self, *args, **kw):
        content = ''
        if self.inside_form:
            content = self.template(*args, **kw)
        else:
            content = self.form_wrapper(*args, **kw)

        if self.fromPublication:
            return content

        zc.resourcelibrary.need('schooltool.table')
        return self.container_wrapper(content=content)


class TableFilter(flourish.viewlet.Viewlet, FilterWidget):
    implements(IFilterWidget)

    before = ("batch", "table")

    template = ViewPageTemplateFile("templates/f_filter.pt")
    title = _("Search")
    legend = _("Search")

    @property
    def source(self):
        return self.manager.source

    @property
    def ignoreRequest(self):
        return self.manager.ignoreRequest

    def filter(self, list):
        if self.ignoreRequest:
            return list
        return FilterWidget.filter(self, list)

    @property
    def script(self):
        return "return ST.table.on_form_submit(${html_id}, this);"


class TableBatch(flourish.viewlet.Viewlet):

    before = ("table", )

    batch = None

    def __init__(self, context, request, view, manager):
        flourish.viewlet.Viewlet.__init__(
            self, context, request, view, manager)

    @property
    def html_id(self):
        return flourish.page.generic_viewlet_html_id(self)

    def __iter__(self):
        return iter(self.batch)

    def update(self):
        if self.manager.prefix:
            self.name = "." + self.manager.prefix
        else:
            self.name = ""
        if self.manager.ignoreRequest:
            start = 0
            size = self.manager.batch_size
        else:
            start = int(self.request.get(
                    'start' + self.name, 0))
            size = int(self.request.get(
                    'size' + self.name, self.manager.batch_size))
        items = self.manager._items
        self.batch = TokenBatch(
            items, size=size, start=start)

    @property
    def needsBatch(self):
        batch = self.batch
        return (batch.size < batch.full_size and batch.needsBatch)

    def extend_token(self, token, **kw):
        token = dict(token)
        script = "return ST.table.on_batch_link('%s', '%s', %d, %d);" % (
            self.manager.html_id,
            self.name, token['start'],
            token['size'])
        token['onclick'] = script
        token.update(kw)
        return token

    @property
    def start(self):
        return self.batch.start

    @property
    def size(self):
        return self.batch.size

    @property
    def length(self):
        return len(self.batch)

    @property
    def full_size(self):
        return self.batch.full_size

    @Lazy
    def previous(self):
        token = self.batch.previous()
        if token is None:
            return None
        return self.extend_token(token, css_class='previous')

    @Lazy
    def next(self):
        token = self.batch.next()
        if token is None:
            return None
        return self.extend_token(token, css_class='next')

    @Lazy
    def show_all(self):
        if self.batch.size >= self.batch.full_size:
            return None
        script = "return ST.table.on_batch_link('%s', '%s', %d, %d);" % (
            self.manager.html_id,
            self.name, 0, self.full_size)
        return {'start': 0,
                'size': self.full_size,
                'items': self.batch.items,
                'onclick': script,
                'css_class': 'all'}

    def tokens(self):
        tokens = []
        for token in self.batch.tokens():
            css_class = "batch_page"
            css_class += token['current'] and ' current' or ''
            token = self.extend_token(token, css_class=css_class)
            tokens.append(token)
        return tokens


class TableTable(flourish.viewlet.Viewlet):

    def render(self, *args, **kw):
        return self.manager.renderTable()


class IndexedTable(IndexedTableFormatter, Table):

    @Lazy
    def filter_widget(self):
        return self.get('filter')

    def makeFormatter(self):
        if self._table_formatter is None:
            return ''
        columns = [IIndexedColumn(c) for c in self._columns]
        formatter = self._table_formatter(
            self.source, self.request, self._items,
            columns=columns,
            batch_start=self.batch.start, batch_size=self.batch.size,
            sort_on=self._sort_on,
            prefix=self.prefix,
            group_by_column=self.group_by_column)
        formatter.html_id = self.html_id
        formatter.view = self
        formatter.cssClasses.update(dict(self.css_classes))
        return formatter

    renderTable = Table.renderTable

    def render(self, *args, **kw):
        return Table.render(self, *args, **kw)


class IndexedTableFilter(TableFilter, IndexedFilterWidget):

    def filter(self, list):
        if self.ignoreRequest:
            return list
        return IndexedFilterWidget.filter(self, list)



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2005 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Tests for pdf testing helpers.
"""

import unittest
import doctest

from reportlab.lib.styles import ParagraphStyle
from reportlab.lib import units
from reportlab import platypus


def buildTableFlowable():
    para_style = ParagraphStyle(name='Test', fontName='Times-Roman')

    data = [
        ['text',
         platypus.paragraph.Paragraph('Text', para_style)],
        [['several', 'items in a cell'],
         platypus.flowables.Image('table_img.png')],
    ]
    return platypus.tables.Table(data)


def buildNestedTables():
    data = [
        ['A table with another table inside!'],
        [buildTableFlowable()]]
    return platypus.tables.LongTable(data)


def buildTestStory():
    para_style = ParagraphStyle(name='Test', fontName='Times-Roman')

    flowables = []
    flowables.append('Some\ntext')

    flowables.append(
        platypus.flowables.KeepInFrame(
            units.inch*2, units.inch, content=[
                platypus.paragraph.Paragraph('Single line', para_style),
                u'unicode text']))
    flowables.append(
        platypus.xpreformatted.PythonPreformatted('print "foo"',
                                                  para_style, bulletText='*'))
    flowables.append(
        platypus.flowables.KeepTogether([
            platypus.paragraph.Paragraph(
                'Multi &amp;\n<b>Line</b>', para_style),
            platypus.xpreformatted.XPreformatted(
                'Text', para_style, bulletText='*'),
            ],
            maxHeight=units.inch))

    flowables.extend([
        platypus.flowables.HRFlowable(),
        platypus.flowables.Image('logo.png', height=units.inch),
        platypus.flowables.ParagraphAndImage(
            platypus.paragraph.Paragraph('Text', para_style),
            platypus.flowables.Image('file.png'),
            xpad=units.inch),
        ])

    flowables.extend([
        platypus.flowables.PageBreak(),
        platypus.flowables.SlowPageBreak(),
        platypus.flowables.CondPageBreak(height=units.inch*2),
        platypus.flowables.Spacer(units.inch*3, units.inch),
        ])

    flowables.append(platypus.flowables.AnchorFlowable('My anchor'))

    # also add some uninteresting flowables
    flowables.append(platypus.flowables.UseUpSpace())
    flowables.append(platypus.flowables.Macro('print "foo"'))

    flowables.append(buildNestedTables())
    return flowables


def doctest_StoryXML():
    r"""Tests for StoryXML document building and printing

        >>> story = buildTestStory()

    StoryXML builds an XML element tree with some some basic flowable
    parameters.

        >>> from schooltool.testing.pdf import StoryXML
        >>> parser = StoryXML(story)

    The XML document can be accessed directly:

        >>> parser.document
        <...ElementTree object ...>

    StoryXML.printXML prints the XML tree.

        >>> parser.printXML()
        <story>
          Some
          text
          <KeepInFrame>
            <Paragraph>Single line</Paragraph>
            unicode text
          </KeepInFrame>
          <PythonPreformatted bulletText="*">print "foo"</PythonPreformatted>
          <KeepTogether>
            <Paragraph>Multi &amp; Line</Paragraph>
            <XPreformatted bulletText="*">Text</XPreformatted>
          </KeepTogether>
          <HRFlowable width="80%" lineWidth="1"
                      spaceBefore="1" spaceAfter="1"
                      hAlign="CENTER" vAlign="BOTTOM"/>
          <Image filename="logo.png" width="None" height="72.0"/>
          <ParagraphAndImage xpad="72.0" ypad="3">
            <Image filename="file.png" width="None" height="None"/>
            <Paragraph>Text</Paragraph>
          </ParagraphAndImage>
          <PageBreak/>
          <SlowPageBreak/>
          <CondPageBreak height="144.0"/>
          <Spacer width="216.0" height="72.0"/>
          <AnchorFlowable name="My anchor"/>
          <LongTable>
          <tr>
            <td>A table with another table inside!</td>
          </tr>
          <tr>
            <td><Table>
              <tr>
                <td>text</td>
                <td><Paragraph>Text</Paragraph></td>
              </tr>
              <tr>
                <td>several
                    items in a cell</td>
                <td><Image filename="table_img.png" width="None" height="None"/></td>
              </tr>
            </Table></td>
            </tr>
          </LongTable>
        </story>

    We can pass an XPath query to printXML.

        >>> parser.printXML('//Image')
        <Image filename="logo.png" width="None" height="72.0"/>
        <Image filename="file.png" width="None" height="None"/>
        <Image filename="table_img.png" width="None" height="None"/>

    Or use SimpleXML.query to obtain a list of results.

        >>> parser.query('//Image')
        ['<Image filename="logo.png" width="None" height="72.0"/>',
         '<Image filename="file.png" width="None" height="None"/>',
         '<Image filename="table_img.png" width="None" height="None"/>']

    """


def test_suite():
    optionflags = (doctest.NORMALIZE_WHITESPACE |
                   doctest.ELLIPSIS |
                   doctest.REPORT_NDIFF)
    return unittest.TestSuite((
        doctest.DocTestSuite(optionflags=optionflags),
        ))

if __name__ == '__main__':
    unittest.main(default='test_suite')



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2013 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Functional selenium tests for schooltool.timetable
"""
import unittest

from schooltool.testing.selenium import collect_ftests
from schooltool.testing.util import format_table
from schooltool.timetable.stesting import timetable_selenium_layer


def print_schedules(browser):
    title_sel = '.container .body > div > h3:not(".done-link")'
    rows = []
    for title in browser.driver.execute_script(
        'return $(arguments[0])', title_sel):
        print title.text
        for table in browser.driver.execute_script(
            'return $(arguments[0]).next().find(arguments[1])',
            title, 'table.timetable'):
            rows.append(
                [th.text
                 for th in browser.driver.execute_script(
                        'return $(arguments[0]).find(arguments[1])',
                        table, 'th.day')])
            for tr in browser.driver.execute_script(
                'return $(arguments[0]).find(arguments[1])',
                table, 'tbody tr'):
                row = []
                for td in browser.driver.execute_script(
                    'return $(arguments[0]).find(arguments[1])',
                    tr, 'td'):
                    row.append(td.text)
                rows.append(row)
    print format_table(rows, header_rows=1)


def test_suite():
    extra_globs = {
        'print_schedules': print_schedules,
        }
    return collect_ftests(layer=timetable_selenium_layer,
                          extra_globs=extra_globs)


if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2011 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
SchoolTool flourish widgets.
"""
import time, datetime
import re
import os
import sys
from cStringIO import StringIO

try:
    import Image
except ImportError:
    from PIL import Image

import zope.formlib.widgets
import zope.datetime
from zope.component import getUtility, adapter, adapts, queryMultiAdapter
from zope.dublincore.interfaces import IZopeDublinCore
from zope.interface import implementer, Interface, implements
from zope.i18n.interfaces import INegotiator
from zope.traversing.browser.absoluteurl import absoluteURL
from zope.publisher.interfaces import NotFound
from zope.publisher.browser import BrowserPage
from zope.publisher.browser import BrowserView
from zope.schema.interfaces import IField

import z3c.form.interfaces
from z3c.form.browser.file import FileWidget
from z3c.form.widget import ComputedWidgetAttribute, FieldWidget
from z3c.form.converter import FileUploadDataConverter
from z3c.form.converter import FormatterValidationError
import zc.resourcelibrary

from schooltool.app.interfaces import ISchoolToolApplication
from schooltool.basicperson.demographics import IDemographicsForm
from schooltool.skin.widgets import CkeditorConfig
from schooltool.skin.widgets import ICkeditorWidget
from schooltool.skin.widgets import CkeditorFormlibWidget
from schooltool.skin.widgets import CkeditorZ3CFormWidget
from schooltool.skin.flourish.resource import ResourceLibrary
from schooltool.skin.flourish.interfaces import IFlourishLayer
from schooltool.skin.flourish.helpers import quoteFilename
from schooltool.common.fields import IImage, ImageFile
from schooltool.common import format_message
from schooltool.common import SchoolToolMessage as _


class JQueryI18nLibrary(ResourceLibrary):

    lib_path = None

    @classmethod
    def configure(cls):
        translation_path = cls.source
        if cls.lib_path is None:
            gdict = sys._getframe(1).f_globals
            lib_path = os.path.dirname(gdict["__file__"])
        full_path = os.path.join(lib_path, translation_path)

        locale_re = re.compile('jquery[.]ui[.]datepicker-'
                               '(?P<locale>[a-zA-Z]+([-_][a-zA-Z]+)?)'
                               '[.]js')

        cls.locales = cls.read_locale_map(full_path, locale_re)

    @classmethod
    def read_locale_map(cls, path, locale_re):
        if not os.path.isdir(path):
            raise ValueError("Not a translation directory", path)
        contents = os.listdir(path)
        locales = {}
        for name in contents:
            m = locale_re.match(name)
            if m is None:
                continue
            locale = m.groupdict()['locale']
            locales[locale] = name
        return locales

    @property
    def included(self):
        negotiator = getUtility(INegotiator)
        lang = negotiator.getLanguage(self.locales, self.request)
        if lang in self.locales:
            return (self.locales[lang], )
        return ()


class FormlibDateWidget(zope.formlib.widgets.DateWidget):
    cssClass="date-field"


Flourish_ckeditor_config = ComputedWidgetAttribute(
    lambda a: CkeditorConfig(288, 160),
    request=IFlourishLayer,
    widget=ICkeditorWidget,
    )


class FlourishCkeditorScriptBase(object):

    @property
    def script(self):
        zc.resourcelibrary.need("ckeditor")
        config = self.config

        app_url = absoluteURL(ISchoolToolApplication(None), self.request)
        config_path = '%s%s' % (
            app_url, config.path)
        skin_path = '%s/@@/schooltool.skin.flourish-ckeditor/' % (
            app_url)
        contents_css_path = '%s%s' % (skin_path, 'contents.css')

        # XXX: using some values that may be not JS safe
        return '''
            <script type="text/javascript" language="JavaScript">
                var skin = "v2,%(skinPath)s";
                if (CKEDITOR.skins == undefined) {
                    skin = "moono";
                }
                var %(variable)s = new CKEDITOR.replace("%(id)s",
                    {
                        height: %(height)s,
                        width: %(width)s,
                        customConfig: "%(customConfigPath)s",
                        skin: skin,
                        contentsCss: "%(contentsCss)s"
                    }
                );
            </script>
            ''' % {
            'id': self.element_id,
            'variable': self.editor_var_name,
            'width': config.width,
            'height': config.height,
            'customConfigPath': config_path,
            'skinPath': skin_path,
            'contentsCss': contents_css_path,
            }


class FlourishCkeditorFormlibWidget(FlourishCkeditorScriptBase,
                                    CkeditorFormlibWidget):

    def __init__(self, *args, **kw):
        super(FlourishCkeditorFormlibWidget, self).__init__(*args, **kw)
        self.config = CkeditorConfig(288, 160)


class FlourishCkeditorZ3CFormWidget(FlourishCkeditorScriptBase,
                                    CkeditorZ3CFormWidget):

    pass


@adapter(IField, IFlourishLayer)
@implementer(z3c.form.interfaces.IFieldWidget)
def FlourishCkeditorFieldWidget(field, request):
    return FieldWidget(field, FlourishCkeditorZ3CFormWidget(request))


def is_required_demo_field(adapter):
    field = adapter.field
    return field.required and field.interface is IDemographicsForm


PromptRequiredDemoField = ComputedWidgetAttribute(is_required_demo_field)


class FileDataURI(BrowserView):

    def __call__(self):
        stream = self.context.open()
        payload = stream.read().encode('base64').replace('\n','')
        stream.close()
        mime = self.context.mimeType
        return 'data:'+mime+';base64,'+payload


class IImageWidget(Interface):
    """Marker for image widgets"""


class ImageWidget(FileWidget):

    implements(IImageWidget)

    @property
    def attribute(self):
        return self.field.__name__

    @property
    def alt(self):
        return self.field.title

    def stored_value(self):
        if self.ignoreContext:
            return None
        dm = queryMultiAdapter((self.context, self.field),
                               z3c.form.interfaces.IDataManager)
        if dm is None:
            return None
        value = dm.query()
        return value


def ImageFieldWidget(field, request):
    return FieldWidget(field, ImageWidget(request))


class ImageDataConverter(FileUploadDataConverter):

    adapts(IImage, IImageWidget)

    def toFieldValue(self, value):
        if value is None or value == '':
            return z3c.form.interfaces.NOT_CHANGED
        if value == 'delete':
            # XXX: delete checkbox was marked
            return None
        data = value.read()
        try:
            image = Image.open(StringIO(data))
            if image.format not in ('JPEG', 'PNG'):
                raise IOError()
        except (IOError,):
            raise FormatterValidationError(
                _('The file uploaded is not a JPEG or PNG image'), value)
        size = len(data)
        if size > self.field.max_file_size:
            msg = _('The image uploaded cannot be larger than ${size} MB')
            raise FormatterValidationError(
                format_message(
                    msg,
                    mapping={'size': '%.2f' % (float(size) / (10**6))}),
                value)
        data = self.processImage(image)
        f = ImageFile()
        self.updateFile(f, data)
        return f

    def processImage(self, image):
        kw = {'quality': 100, 'filter': Image.ANTIALIAS}
        result = image
        if (self.field.size is not None and
            image.size != self.field.size):
            size = self.getMaxSize(image)
            image.thumbnail(size, kw['filter'])
            if image.size != self.field.size or image.mode == 'RGBA':
                result = Image.new('RGB', self.field.size, (255, 255, 255))
                left = int(round((self.field.size[0] - image.size[0])/float(2)))
                up = int(round((self.field.size[1] - image.size[1])/float(2)))
                mask = None
                if image.mode == 'RGBA':
                    mask = image
                result.paste(image, (left, up), mask)
        f = StringIO()
        result.save(f, self.field.format, **kw)
        return f.getvalue()

    def updateFile(self, ob, data):
        ob.mimeType = Image.MIME[self.field.format]
        w = ob.open("w")
        w.write(data)
        w.close()

    def getMaxSize(self, image):
        image_size = image.size
        desired_size = self.field.size
        x_ratio = float(desired_size[0])/image_size[0]
        y_ratio = float(desired_size[1])/image_size[1]
        if x_ratio < y_ratio:
            new_size = (round(x_ratio*image_size[0]),
                        round(x_ratio*image_size[1]))
        else:
            new_size = (round(y_ratio*image_size[0]),
                        round(y_ratio*image_size[1]))
        return tuple(map(int, new_size))


class ImageView(BrowserPage):

    attribute = None

    def renderImage(self, image):
        self.request.response.setHeader('Content-Type', image.mimeType)
        self.request.response.setHeader('Content-Length', image.size)
        try:
            modified = IZopeDublinCore(self.context).modified
        except TypeError:
            modified=None
        if modified is not None and isinstance(modified, datetime.datetime):
            header= self.request.getHeader('If-Modified-Since', None)
            lmt = long(time.mktime(modified.timetuple()))
            if header is not None:
                header = header.split(';')[0]
                try:
                    mod_since=long(time(header))
                except:
                    mod_since=None
                if mod_since is not None:
                    if lmt <= mod_since:
                        self.request.response.setStatus(304)
                        return ''
            self.request.response.setHeader(
                'Last-Modified', zope.datetime.rfc1123_date(lmt))
        result = image.openDetached()
        return result

    @property
    def image(self):
        if self.attribute:
            try:
                image = getattr(self.context, self.attribute)
            except AttributeError:
                raise NotFound(self.context, self.attribute, self.request)
        else:
            image = self.context
        return image

    def __call__(self):
        image = self.image
        if image is None:
            return ''

        result = self.renderImage(image)
        return result


class DownloadFile(BrowserView):

    inline = False

    def __call__(self):
        response = self.request.response
        mime = self.context.mimeType
        if mime:
            response.setHeader('Content-Type', mime)
        response.setHeader('Content-Length', self.context.size)
        if self.inline:
            # We don't really accept ranges, but Acrobat Reader will not show the
            # report in the browser page if this header is not provided.
            response.setHeader('Accept-Ranges', 'bytes')
        disposition = self.inline and 'inline' or 'attachment'
        filename = self.context.__name__
        quoted_filename = quoteFilename(filename)
        if quoted_filename:
            disposition += '; filename="%s"' % quoted_filename
        response.setHeader('Content-Disposition', disposition)
        return self.context.openDetached()





#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2007 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Sources and vocabularies for form fields.
"""
from zope.interface import implements
from zope.interface import implementer
from zope.component import adapter

from schooltool.app.interfaces import ISchoolToolApplication
from schooltool.app.utils import vocabulary
from schooltool.person.interfaces import IPerson
from schooltool.group.interfaces import IGroupContainer
from schooltool.group.interfaces import IGroup

from schooltool.basicperson.browser.person import PersonTerm
from schooltool.basicperson.browser.person import GroupTerm
from schooltool.basicperson.interfaces import IGroupVocabulary
from schooltool.basicperson.interfaces import IBasicPersonVocabulary
from schooltool.basicperson.interfaces import IDemographicsFields
from schooltool.basicperson.interfaces import IFieldDescription
from schooltool.basicperson.interfaces import IFieldFilterVocabulary

from schooltool.common import SchoolToolMessage as _


class GroupVocabulary(object):
    implements(IGroupVocabulary)

    def __init__(self, context):
        self.context = context

    def __len__(self):
        return len(self.context.groups)

    def __contains__(self, value):
        return True

    def __iter__(self):
        if IPerson.providedBy(self.context):
            groups = [group for group in self.context.groups
                      if IGroup.providedBy(group)]
        else:
            groups = list(IGroupContainer(ISchoolToolApplication(None), {}).values())

        for group in sorted(groups, key=lambda g: g.__name__):
            yield GroupTerm(group)

    def getTermByToken(self, token):
        gc = IGroupContainer(ISchoolToolApplication(None))
        if gc is None:
            raise LookupError(token)
        if token not in gc:
            raise LookupError(token)
        return GroupTerm(gc[token])

    def getTerm(self, value):
        return GroupTerm(value)


def groupVocabularyFactory():
    return GroupVocabulary


class AdvisorVocabulary(object):
    implements(IBasicPersonVocabulary)

    def __init__(self, context):
        self.context = context

    def __contains__(self, value):
        return True

    def __len__(self):
        return len(self.context.advisors)

    def __iter__(self):
        gc = IGroupContainer(ISchoolToolApplication(None), None)
        if gc is None:
            return
        persons = gc['teachers'].members
        for person in sorted(persons, key=lambda p: p.__name__):
            yield PersonTerm(person)

    def getTermByToken(self, token):
        app = ISchoolToolApplication(None)
        pc = app['persons']
        if token not in pc:
            raise LookupError(token)
        return PersonTerm(pc[token])

    def getTerm(self, value):
        return PersonTerm(value)


def advisorVocabularyFactory():
    return AdvisorVocabulary


def LimitKeyVocabFactory():
    return IFieldFilterVocabulary


@adapter(IFieldDescription)
@implementer(IFieldFilterVocabulary)
def getLimitKeyVocabularyForFieldDescription(field_description):
    field_container = field_description.__parent__
    return IFieldFilterVocabulary(field_container)


@adapter(IDemographicsFields)
@implementer(IFieldFilterVocabulary)
def getLimitKeyVocabularyForPersonFields(person_field_description_container):
    return vocabulary([
        ('students', _('Students')),
        ('teachers', _('Teachers')),
        ('administrators', _('Administrators')),
        ])




#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2005 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Timetable setup wizard for SchoolTool.

This module implements a the workflow described by Tom Hoffman on Jun 9 2005
on the schooltool mailing list, in ttschema-wireframes.pdf[1].

    [1] http://lists.schooltool.org/pipermail/schooltool/2005-June/001347.html

It has been slightly modified since.  You can find an up-to-date diagram
of the workflow in src/schooltool/app/browser/ftests/images/

The workflow is as follows:

    1. "New timetable"

       (The user enters a name.)

    2. "Does your school's timetable cycle use days of the week, or a rotating
        cycle?"

        Skip to step 4 if days of the week was chosen.

    3. "Enter names of days in cycle:"

        (The user enters a list of day names in a single textarea.)

    4. "Do classes begin and end at the same time each day in your school's
        timetable?"

        Continue with step 5 if "yes".
        Skip to step 6 if "no", and "cycle" was chosen in step 2.
        Skip to step 7 if "no", and "days of the week" was chosen in step 2.

    5. "Enter start and end times for each slot"

        (The user enters a list of start and end times in a single textarea.)

        Jump to step 9.

    6. "Do the start and end times vary based on the day of the week, or the
        day in the cycle?"

        Continue with step 7 if "days of week".
        Continue with step 8 if "cycle".

    7. "Enter the start and end times of each slot on each day:"

        (The user sees 5 columns of text lines, with five buttons that let him
        add an extra slot for each column.)

        Jump to step 9.

    8. "Enter the start and end times of each slot on each day:"

        (The user sees N columns of text lines, with N buttons that let him
        add an extra slot for each column.)

    9. "Do periods have names or are they simply designated by time?"

        Skip to step 14 if periods are simply designated by time.

    10. "Enter names of the periods:"

        (The user enters a list of periods in a single textarea.)

    11. "Is the sequence of periods each day the same or different?"

        Skip to step 13 if it is different

    12. "Put the periods in order for each day:"

        (The user sees a list of drop-downs.)

        Jump to step 14.

    13. "Put the periods in order:"

        (The user sees a grid of drop-downs.)

    14. "Does your school have a homeroom period?"

        Skip to step 16 if it doesn't.

    15. "Indicate which period is the homeroom period."

        (The user sees a list of drop-downs.)

    16. The timetable is created.


The shortest path through this workflow contains 7 steps, the longest contains
13 steps.

Step 16 needs the following data:

    - title (determined in step 1)
    - timetable model factory (determined in step 2)
    - a list of day IDs (determined in step 3)
    - a list of periods for each day ID
    - a list of day templates (weekday -> period_id -> start time & duration)

"""
import os

from zope.interface import Interface
from zope.schema import TextLine, Text, getFieldNamesInOrder
from zope.i18n import translate
from zope.app.form.utility import setUpWidgets
from zope.app.form.utility import getWidgetsData
from zope.app.form.interfaces import IInputWidget
from zope.app.form.interfaces import WidgetsError
from zope.browserpage.viewpagetemplatefile import ViewPageTemplateFile
from zope.browserpage.viewpagetemplatefile import BoundPageTemplate
from zope.container.interfaces import INameChooser
from zope.container.contained import containedEvent
from zope.event import notify
from zope.publisher.browser import BrowserView
from zope.session.interfaces import ISession
from zope.traversing.browser.absoluteurl import absoluteURL

import schooltool.skin.flourish.content
import schooltool.skin.flourish.form
import schooltool.skin.flourish.page
from schooltool.app.browser.cal import day_of_week_names
from schooltool.app.interfaces import ISchoolToolApplication
from schooltool.app.interfaces import IApplicationPreferences
from schooltool.common import parse_time_range, format_time_range
from schooltool.common.inlinept import InlineViewPageTemplate
from schooltool.skin import flourish
from schooltool.skin.flourish.tal import JSONDecoder
from schooltool.schoolyear.interfaces import ISchoolYearContainer
from schooltool.timetable.interfaces import IHaveTimetables
from schooltool.timetable.daytemplates import WeekDayTemplates
from schooltool.timetable.daytemplates import SchoolDayTemplates
from schooltool.timetable.daytemplates import DayTemplate
from schooltool.timetable.daytemplates import TimeSlot
from schooltool.timetable.schedule import Period
from schooltool.timetable.timetable import Timetable

from schooltool.common import SchoolToolMessage as _


def getSessionData(view):
    """Return the data container stored in the session."""
    return ISession(view.request)['schooltool.ttwizard']

#
# Abstract step classes
#

class Step(BrowserView):
    """A step, one of many.

    Each step has three important methods:

        `__call__` renders the page.

        `update` processes the form, and then either stores the data in the
        session, and returns True if the form was submitted correctly, or
        returns False if it wasn't.

        `next` returns the next step.

    """
    getSessionData = getSessionData

    @property
    def label(self):
        session = self.getSessionData()
        return _("Timetable $title",
                 mapping={'title': session['title']})



class ChoiceStep(Step):
    """A step that requires the user to make a choice.

    Subclasses should provide three attributes:

        `question` -- question text

        `choices` -- a list of tuples (choice_value, choice_text)

        `key` -- name of the session dict key that will store the value.

    They should also define the `next` method.
    """

    __call__ = ViewPageTemplateFile("templates/ttwizard_choice.pt")

    def update(self):
        session = self.getSessionData()
        for n, (value, text) in enumerate(self.choices):
            if 'NEXT.%d' % n in self.request:
                session[self.key] = value
                return True
        return False


class FormStep(Step):
    """A step that presents a form.

    Subclasses should provide a `schema` attribute.

    They can override the `description` attribute and specify some informative
    text to be displayed above the form.

    They should also define the `update` and `next` methods.
    """

    __call__ = ViewPageTemplateFile("templates/ttwizard_form.pt")

    description = None

    error = None

    def __init__(self, context, request):
        BrowserView.__init__(self, context, request)
        setUpWidgets(self, self.schema, IInputWidget)

    def widgets(self):
        return [getattr(self, name + '_widget')
                for name in getFieldNamesInOrder(self.schema)]


#
# Concrete wizard steps
#

class FirstStep(FormStep):
    """First step: enter the title for the new timetable."""

    __name__ = 'first_step'

    __call__ = ViewPageTemplateFile("templates/ttwizard.pt")

    label = _('A new timetable')

    class schema(Interface):
        title = TextLine(title=_("Title"), default=u"default")

    def update(self):
        try:
            data = getWidgetsData(self, self.schema)
        except WidgetsError:
            return False
        session = self.getSessionData()
        session['title'] = data['title']
        return True

    def next(self):
        return CycleStep


class CycleStep(ChoiceStep):
    """A step for choosing the timetable cycle."""

    __name__ = 'cycle_step'
    key = 'cycle'

    question = _("Does your school's timetable cycle use days of the week,"
                 " or a rotating cycle?")

    choices = (('weekly',   _("Days of the week")),
               ('rotating', _("Rotating cycle")))

    def next(self):
        session = self.getSessionData()
        if session['cycle'] == 'weekly':
            return IndependentDaysStep
        else:
            return DayEntryStep

    def update(self):
        success = ChoiceStep.update(self)
        session = self.getSessionData()
        if success and session['cycle'] == 'weekly':
            weekday_names = [translate(day_of_week_names[i],
                                       context=self.request)
                             for i in range(7)]
            session['day_names'] = weekday_names
        session['time_model'] = session['cycle']
        return success


class DayEntryStep(FormStep):
    """A step for entering names of days."""

    __name__ = 'day_entry_step'
    description = _("Enter names of days in cycle, one per line.")

    class schema(Interface):
        days = Text(required=False)

    def update(self):
        try:
            data = getWidgetsData(self, self.schema)
        except WidgetsError, e:
            return False
        day_names = parse_name_list(data.get('days') or '')
        if not day_names:
            self.error = _("Please enter at least one day name.")
            return False

        seen = set()
        for day in day_names:
            if day in seen:
                self.error = _("Please make sure the day names are unique.")
                return False
            seen.add(day)
        session = self.getSessionData()
        session['day_names'] = day_names
        return True

    def next(self):
        return IndependentDaysStep


class IndependentDaysStep(ChoiceStep):
    """A step for choosing if all period times are the same each day."""

    __name__ = 'independent_days_step'
    key = 'similar_days'

    question = _("Do classes begin and end at the same time each day in"
                 " your school's timetable?")

    choices = ((True,  _("Same time each day")),
               (False, _("Different times")))

    def next(self):
        session = self.getSessionData()
        if session['similar_days']:
            return SimpleSlotEntryStep
        else:
            if session['cycle'] == 'weekly':
                return WeeklySlotEntryStep
            else:
                return SequentialModelStep


class SequentialModelStep(ChoiceStep):
    """Step for choosing if start and end times vay based on the day of
    the week or the day in the cycle."""

    __name__ = 'sequential_model_step'
    key = 'time_model'

    question = _("Do start and end times vary based on the day of the week"
                 " (Monday - Friday) or the day in the cycle?")

    choices = [('weekly', _("Day of week")),
               ('rotating', _("Day in cycle"))]

    def next(self):
        session = self.getSessionData()
        if session['time_model'] == 'weekly':
            return WeeklySlotEntryStep
        else:
            return RotatingSlotEntryStep


def parse_name_list(names):
    r"""Parse a multi-line string into a list of names.

    One day name per line.  Empty lines are ignored.  Extra spaces
    are stripped.

        >>> parse_name_list(u'Day 1\n Day 2\nDay 3 \n\n\n Day 4 ')
        [u'Day 1', u'Day 2', u'Day 3', u'Day 4']
        >>> parse_name_list(u'  \n\n ')
        []
        >>> parse_name_list(u'')
        []

    """
    return [s.strip() for s in names.splitlines() if s.strip()]


def parse_time_range_list(times):
    r"""Parse a multi-line string into a list of time slots.

    One slot per line (HH:MM - HH:MM).  Empty lines are ignored.  Extra
    spaces are stripped.

        >>> parse_time_range_list(u' 9:30 - 10:25 \n \n 12:30 - 13:25 \n')
        [(datetime.time(9, 30), datetime.timedelta(0, 3300)),
         (datetime.time(12, 30), datetime.timedelta(0, 3300))]

        >>> parse_time_range_list(u'  \n\n ')
        []
        >>> parse_time_range_list(u'')
        []

        >>> parse_time_range_list(u'9:30-12:20\nbad value\nanother bad value')
        Traceback (most recent call last):
          ...
        ValueError: bad value

    """
    result = []
    for line in times.splitlines():
        line = line.strip()
        if line:
            try:
                result.append(parse_time_range(line))
            except ValueError:
                raise ValueError(line)
    return result


def format_time_range_list(times):
    r"""Format a multi-line string from a list of time slots.

    One slot per line (HH:MM - HH:MM).  Empty lines are ignored.  Extra
    spaces are stripped.

        >>> import datetime
        >>> times = [(datetime.time(9, 30), datetime.timedelta(0, 3300)),
        ...          (datetime.time(12, 30), datetime.timedelta(0, 3300))]

        >>> print format_time_range_list(times)
        09:30-10:25
        12:30-13:25

    """
    lines = []
    for tr_time, tr_delta in times:
        lines.append(format_time_range(tr_time, tr_delta))

    return '\n'.join(lines)


class SimpleSlotEntryStep(FormStep):
    """A step for entering times for classes.

    This step is used when the times are the same in each day.
    """

    __name__ = 'simple_slot_entry_step'

    description = _("Enter start and end times for each slot,"
                    " one slot (HH:MM - HH:MM) per line.")

    class schema(Interface):
        times = Text(default=u"9:30-10:25\n10:30-11:25", required=False)

    def update(self):
        try:
            data = getWidgetsData(self, self.schema)
        except WidgetsError, e:
            return False
        try:
            times = parse_time_range_list(data.get('times') or '')
        except ValueError, e:
            self.error = _("Not a valid time slot: $slot.",
                           mapping={'slot': unicode(e.args[0])})
            return False
        if not times:
            self.error = _("Please enter at least one time slot.")
            return False
        session = self.getSessionData()
        if session.get('similar_days') and session['cycle'] == 'weekly':
            session['time_slots'] = [times] * 5 + [[]] * 2
        else:
            num_days = len(session['day_names'])
            session['time_slots'] = [times] * num_days
        return True

    def next(self):
        return NamedPeriodsStep


class RotatingSlotEntryStep(Step):
    """Step for entering start and end times of slots in each day.

    This step is taken when the start/end times are different for each day,
    the rotating cycle is chosen in the CycleStep and 'day in the cycle' is
    chosen in the SequentialModelStep.
    """

    __name__ = 'rotating_slot_entry_step'
    __call__ = ViewPageTemplateFile("templates/ttwizard_slottimes.pt")

    description = _("Enter start and end times for each slot on each day,"
                    " one slot (HH:MM - HH:MM) per line.")

    error = None
    example_intervals = '8:00 - 8:45\n9:05 - 9:50\n'

    def days(self):
        """Return a list of day ids/values for the view"""
        for day_number in range(len(self.dayNames())):
            day_id = "times.%s" % day_number
            default = ""
            if day_number < 5:
                default = self.example_intervals

            value = self.request.get(day_id, default)
            yield {'id': day_id,
                   'value': value}

    def dayNames(self):
        """Return the list of day names."""
        return self.getSessionData()['day_names']

    def update(self):
        result = []
        for i, day_name in enumerate(self.dayNames()):
            s = self.request.form.get('times.%d' % i, '')
            try:
                times = parse_time_range_list(s)
            except ValueError, e:
                self.error = _("Not a valid time slot: $slot.",
                               mapping={'slot': unicode(e.args[0])})
                return False
            result.append(times)

        session = self.getSessionData()
        session['time_slots'] = result
        return True

    def next(self):
        return NamedPeriodsStep


class WeeklySlotEntryStep(RotatingSlotEntryStep):
    """Step for entering start and end times of slots in each day.

    This step is taken when the start/end times are different for each day,
    and the weekly cycle is chosen in the CycleStep or in the
    SequentialModelStep.
    """

    __name__ = 'weekly_slot_entry_step'

    def dayNames(self):
        """Return the list of day names."""
        return [translate(day_of_week_names[i], context=self.request)
                for i in range(7)]

    def update(self):
        """Store the slots in the session.

        If a rotating timetable cycle is selected and slots vary based on day
        of week, each day must have the same number of slots.
        """
        result = RotatingSlotEntryStep.update(self)
        session = self.getSessionData()
        if (result and session['cycle'] == 'rotating'
            and session.get('time_model') == 'weekly'):
            slots = session['time_slots']
            assert len(slots) == 7 # Monday - Sunday
            num_slots = len(slots[0])
        return result


class NamedPeriodsStep(ChoiceStep):
    """A step for choosing if periods have names or are designated by time"""

    __name__ = 'named_periods_step'
    key = 'named_periods'

    question = _("Do periods have names or are they simply"
                 " designated by time?")

    @property
    def choices(self):
        session = self.getSessionData()
        if session['cycle'] == 'rotating':
            choices = ((True,  _("Have names")),)
        else:
            choices = ((True,  _("Have names")),
                       (False, _("Designated by time")))
        return choices

    def next(self):
        session = self.getSessionData()
        if session['named_periods']:
            return PeriodNamesStep
        else:
            periods_order = default_period_names(session['time_slots'])
            session['periods_order'] = periods_order
            return HomeroomStep


def default_period_names(time_slots):
    """Derive period names and order from time slots.

        >>> from datetime import time, timedelta
        >>> periods = [(time(9, 0), timedelta(minutes=50)),
        ...            (time(12, 35), timedelta(minutes=50)),
        ...            (time(14, 15), timedelta(minutes=55))]
        >>> periods2 = [(time(9, 10), timedelta(minutes=50)),
        ...             (time(12, 35), timedelta(minutes=50))]
        >>> default_period_names([periods] * 3 + [periods2])
        [['09:00-09:50', '12:35-13:25', '14:15-15:10'],
         ['09:00-09:50', '12:35-13:25', '14:15-15:10'],
         ['09:00-09:50', '12:35-13:25', '14:15-15:10'],
         ['09:10-10:00', '12:35-13:25']]

    """
    return [[format_time_range(tstart, duration)
             for tstart, duration in slots]
            for slots in time_slots]


class PeriodNamesStep(FormStep):
    """A step for entering names of periods"""

    __name__ = 'period_names_step'
    description = _("Enter names of periods, one per line.")

    class schema(Interface):
        periods = Text(required=False)

    def update(self):
        try:
            data = getWidgetsData(self, self.schema)
        except WidgetsError, e:
            return False
        periods = parse_name_list(data.get('periods') or '')
        min_periods = self.requiredPeriods()
        if len(periods) < min_periods:
            self.error = _("Please enter at least $number periods.",
                           mapping={'number': min_periods})
            return False
        seen = set()
        for period in periods:
            if period in seen:
                self.error = _("Please make sure the period names are unique.")
                return False
            seen.add(period)
        session = self.getSessionData()
        session['period_names'] = periods
        return True

    def requiredPeriods(self):
        """Return the maximum number of slots there can be on a day."""
        session = self.getSessionData()
        times = session['time_slots']
        return max(map(len, times))

    def next(self):
        return PeriodSequenceSameStep


class PeriodSequenceSameStep(ChoiceStep):
    """A step for choosing whether periods are the same on all days"""

    __name__ = 'period_sequence_same_step'
    key = 'periods_same'

    question = _("Is the sequence of periods each day the same or different?")

    choices = ((True,  _("Same")),
               (False, _("Different")))

    def next(self):
        session = self.getSessionData()
        if session[self.key]:
            return PeriodOrderSimple
        else:
            return PeriodOrderComplex


class PeriodOrderSimple(Step):
    """Step to put periods in order if all days are the same."""

    __name__ = 'period_order_simple_step'

    __call__ = ViewPageTemplateFile('templates/ttwizard_period_order1.pt')

    description = _('Please put the periods in order:')

    error = None

    def periods(self):
        return self.getSessionData()['period_names']

    def numPeriods(self):
        return max([len(day) for day in self.getSessionData()['time_slots']])

    def update(self):
        result = []
        periods = self.getSessionData()['period_names']
        for i in range(self.numPeriods()):
            name = 'period_%d' % i
            if name not in self.request:
                self.error = _('Please provide all periods.')
                return False
            result.append(self.request[name])

        # Validate that all periods are selected
        seen = set()
        errors = set()
        for period in result:
            if period not in seen:
                seen.add(period)
            else:
                errors.add(period)
        if errors:
            self.error = _('The following periods were selected more'
                           ' than once: $periods',
                           mapping={'periods': ', '.join(errors)})
            return False

        day_names = self.getSessionData()['day_names']
        self.getSessionData()['periods_order'] = [result] * len(day_names)
        return True

    def next(self):
        return HomeroomStep


class PeriodOrderComplex(Step):
    """Step to put periods in order if order is different on different days"""

    __name__ = 'period_order_complex_step'
    __call__ = ViewPageTemplateFile('templates/ttwizard_period_order2.pt')

    description = _('Please put the periods in order for each day:')

    error = None

    def periods(self):
        return self.getSessionData()['period_names']

    def days(self):
        session = self.getSessionData()
        return session['day_names']

    def numSlots(self):
        session = self.getSessionData()
        if (session['cycle'] != session.get('time_model')):
            np = len(self.periods())
            return [np for day in self.getSessionData()['day_names']]
        return [len(day) for day in self.getSessionData()['time_slots']]

    def update(self):
        result = []
        periods = self.periods()
        numSlots = self.numSlots()
        for i in range(len(self.days())):
            day = []
            for j in range(numSlots[i]):
                name = 'period_%d_%d' % (i, j)
                if name not in self.request:
                    self.error = _('Please provide all periods.')
                    return False
                day.append(self.request[name])
            result.append(day)

        # Validate that all periods are selected
        errors = set()
        for i, day in enumerate(self.days()):
            seen = set()
            for period in result[i]:
                if period not in seen:
                    seen.add(period)
                else:
                    error = _("$period on day $day",
                              mapping={'period': period, 'day': self.days()[i]})
                    errors.add(translate(error, context=self.request))
        if errors:
            self.error = _('The following periods were selected more'
                           ' than once: $periods',
                           mapping={'periods': ', '.join(errors)})
            return False

        self.getSessionData()['periods_order'] = result
        return True

    def next(self):
        return HomeroomStep


class HomeroomStep(ChoiceStep):
    """A step for choosing whether the school has homeroom periods."""

    __name__ = 'homeroom_step'
    key = 'homeroom'

    question = _("Do you check student attendance for the day in a homeroom"
                 " period or equivalent?")

    choices = ((True,  _("Yes")),
               (False, _("No")))

    def next(self):
        session = self.getSessionData()
        if session['homeroom']:
            return HomeroomPeriodsStep
        else:
            return FinalStep


class HomeroomPeriodsStep(Step):
    """Step to indicate the homeroom period for each day."""

    __name__ = 'homeroom_periods_step'
    __call__ = ViewPageTemplateFile('templates/ttwizard_homeroom.pt')

    description = _('Please indicate the homeroom period(s) for each day:')

    error = None

    def periodsInOrder(self):
        return self.getSessionData()['periods_order']

    def days(self):
        session = self.getSessionData()
        return session['day_names']

    def update(self):
        result = []
        for daynr, periods in enumerate(self.periodsInOrder()):
            homeroom_period_ids = []
            for period in periods:
                period_id = 'homeroom_%d_%s' % (daynr, period)
                if period_id in self.request:
                    homeroom_period_ids.append(period)
            result.append(homeroom_period_ids)
        self.getSessionData()['homeroom_periods'] = result
        return True

    def next(self):
        return FinalStep


class FinalStep(Step):
    """Final step: create the timetable."""

    __name__ = 'final_step'

    def __call__(self):
        self.createAndAdd()
        self.request.response.redirect(
            absoluteURL(self.context, self.request))

    def createAndAdd(self):
        timetable = self.createTimetable()
        self.add(timetable)
        self.setUpTimetable(timetable)
        return timetable

    def update(self):
        return True

    def next(self):
        return FirstStep

    def createTimetable(self):
        session = self.getSessionData()
        app = ISchoolToolApplication(None)
        tzname = IApplicationPreferences(app).timezone

        # XXX: quick fix for date range
        owner = IHaveTimetables(self.context)
        first, last = owner.first, owner.last

        timetable = Timetable(
            first, last,
            title=session['title'],
            timezone=tzname)
        return timetable

    def addTimeSlotTemplates(self, day_template_schedule, days):
        for key, title, time_slots in days:
            template = DayTemplate(title=title)
            day_template_schedule.templates[key] = template
            name_chooser = INameChooser(template)
            for tstart, duration in time_slots:
                timeslot = TimeSlot(tstart, duration, activity_type=None)
                key = name_chooser.chooseName('', timeslot)
                template[key] = timeslot

    def createTimeSlots(self, timetable, model, day_titles, time_slots):
        if model == 'weekly':
            timetable.time_slots, object_event = containedEvent(
                WeekDayTemplates(), timetable, 'time_slots')
            days = [
                (unicode(n),
                 translate(day_of_week_names[n], context=self.request),
                 time_slot)
                for n, time_slot in zip(range(7), time_slots)]
        elif model == 'rotating':
            timetable.time_slots, object_event = containedEvent(
                SchoolDayTemplates(), timetable, 'time_slots')
            # An UI limitations at the time of writing
            time_slots = time_slots[:len(day_titles)]
            day_ids = [unicode(n) for n in range(len(day_titles))]
            days = zip(day_ids, day_titles, time_slots)
        else:
            raise NotImplementedError()

        notify(object_event)
        timetable.time_slots.initTemplates()

        self.addTimeSlotTemplates(timetable.time_slots, days)

    def addPeriodTemplates(self, day_template_schedule, days):
        for day_key, day_title, periods in days:
            template = DayTemplate(title=day_title)
            day_template_schedule.templates[day_key] = template

            name_chooser = INameChooser(template)
            for period_title, activity_type in periods:
                period = Period(title=period_title,
                                activity_type=activity_type)
                key = name_chooser.chooseName('', period)
                template[key] = period

    def createPeriods(self, timetable, model, day_titles,
                      period_names, homeroom_periods):
        if model == 'weekly':
            timetable.periods, object_event = containedEvent(
                WeekDayTemplates(), timetable, 'periods')
            day_ids = [unicode(n) for n in range(min(len(day_titles), 7))]
        elif model == 'rotating':
            timetable.periods, object_event = containedEvent(
                SchoolDayTemplates(), timetable, 'periods')
            day_ids = [unicode(n) for n in range(len(day_titles))]
        else:
            raise NotImplementedError()

        notify(object_event)
        timetable.periods.initTemplates()

        periods = []
        for titles, homeroom_titles in zip(period_names, homeroom_periods):
            periods.append(
                [(title, title in homeroom_titles and 'homeroom' or 'lesson')
                 for title in titles])

        days = zip(day_ids, day_titles, periods)

        self.addPeriodTemplates(timetable.periods, days)

    def setUpTimetable(self, timetable):
        session = self.getSessionData()

        # XXX: would be nice to set theese properly in previous steps
        periods_model = session['cycle']
        time_model = session.get('time_model', 'rotating')
        if periods_model == 'weekly':
            time_model = 'weekly'
        if periods_model == 'rotating' and session['similar_days']:
            time_model = 'rotating'

        assert time_model in ('weekly', 'rotating')
        assert periods_model in ('weekly', 'rotating')

        # An UI limitations at the time of writing
        assert not(time_model == 'rotating' and periods_model != 'rotating')

        day_titles = session['day_names']

        # set time slots
        time_slots = session['time_slots']
        self.createTimeSlots(timetable, time_model, day_titles, time_slots)

        # set periods
        period_names = session['periods_order']
        if session['homeroom']:
            homeroom_periods = session['homeroom_periods']
        else:
            homeroom_periods = [tuple() for day in day_titles]

        self.createPeriods(timetable, periods_model, day_titles,
                           period_names, homeroom_periods)

    def add(self, timetable):
        """Add the timetable to self.context."""
        nameChooser = INameChooser(self.context)
        key = nameChooser.chooseName('', timetable)
        self.context[key] = timetable


#
# The wizard itself
#

class TimetableWizard(BrowserView):
    """View for defining a new timetable."""

    getSessionData = getSessionData

    def getLastStep(self):
        session = self.getSessionData()
        step_class = session.get('last_step', FirstStep)
        step = step_class(self.context, self.request)
        return step

    def rememberLastStep(self, step):
        session = self.getSessionData()
        session['last_step'] = step.__class__

    def __call__(self):
        if 'CANCEL' in self.request:
            self.rememberLastStep(FirstStep(self.context, self.request))
            self.request.response.redirect(
                    absoluteURL(self.context, self.request))
            return
        current_step = self.getLastStep()
        if current_step.update():
            next = current_step.next()
            current_step = next(self.context, self.request)
        self.rememberLastStep(current_step)
        return current_step()


class FlourishTimetableWizard(flourish.page.WideContainerPage, TimetableWizard):

    _state = None
    _session = None
    current_step = None

    def getLastStep(self):
        raise NotImplemented('No longer used.')

    def rememberLastStep(self, step):
        raise NotImplemented('No longer used.')

    @property
    def state(self):
        if self._state is None:
            if 'viewstate' in self.request:
                decoder = JSONDecoder()
                self._state = decoder.decode(self.request['viewstate'])
                slots = self._state.get('time_slots')
                if slots is not None:
                    self._state['time_slots'] = [
                        parse_time_range_list(day) for day in slots]
            else:
                self._state = {}
        if not self._state.get('steps'):
            self._state['steps'] = []
        return self._state

    @property
    def json_state(self):
        state = self.state
        slots = state.get('time_slots')
        if slots is not None:
            state['time_slots'] = [
                format_time_range_list(day) for day in slots]
        return state

    def queryStep(self, name):
        return flourish.content.queryContentProvider(
            self.context, self.request, self, name)

    def nextURL(self):
        link = flourish.content.queryContentProvider(
            self.context, self.request, self, 'done_link')
        if link is not None:
            return link.url
        return absoluteURL(self.context, self.request)

    def update(self):
        flourish.page.WideContainerPage.update(self)

        if 'CANCEL' in self.request:
            self.request.response.redirect(self.nextURL())
            return

        steps = self.state['steps']

        if 'STEP_BACK' in self.request:
            if len(steps) > 1:
                steps.pop()
            self.current_step = self.queryStep(steps[-1])
            return

        if not steps:
            first_step_name = FirstStep.__dict__['__name__']
            self.current_step = self.queryStep(first_step_name)
            steps.append(self.current_step.__name__)
            return

        self.current_step = self.queryStep(steps[-1])

        step = self.current_step
        step.update()

        if step.finished:
            next_step = step.next()
            self.current_step = next_step
            steps.append(next_step.__name__)


def flourishStep(base, auto_template=True):

    template = None
    if auto_template and isinstance:
        if isinstance(base.__call__, BoundPageTemplate):
            filename = os.path.split(base.__call__.filename)[-1]
            template = os.path.join('templates', 'f_%s' % filename)

    def getSessionData(self):
        return self.view.state
    def update(self):
        self.finished = base.update(self)
    def next(self):
        cls = base.next(self)
        next_step_name = cls.__dict__['__name__']
        return self.view.queryStep(next_step_name)
    class_dict = {
        'getSessionData': lambda self: self.view.state,
        'update': update,
        'next': next,
        }
    if template is not None:
        class_dict['template'] = ViewPageTemplateFile(template)
    new_class = type('Flourish%s' % base.__name__,
                     (flourish.content.ContentProvider, base),
                     class_dict)
    return new_class


FlourishFirstStep = flourishStep(FirstStep, auto_template=False)
FlourishCycleStep = flourishStep(CycleStep)
FlourishDayEntryStep = flourishStep(DayEntryStep)
FlourishIndependentDaysStep = flourishStep(IndependentDaysStep)
FlourishSequentialModelStep = flourishStep(SequentialModelStep)
FlourishSimpleSlotEntryStep = flourishStep(SimpleSlotEntryStep)
FlourishRotatingSlotEntryStep = flourishStep(RotatingSlotEntryStep)
FlourishWeeklySlotEntryStep = flourishStep(WeeklySlotEntryStep)
FlourishNamedPeriodsStep = flourishStep(NamedPeriodsStep)
FlourishPeriodNamesStep = flourishStep(PeriodNamesStep)
FlourishPeriodSequenceSameStep = flourishStep(PeriodSequenceSameStep)
FlourishPeriodOrderSimple = flourishStep(PeriodOrderSimple)
FlourishPeriodOrderComplex = flourishStep(PeriodOrderComplex)
FlourishHomeroomStep = flourishStep(HomeroomStep)
FlourishHomeroomPeriodsStep = flourishStep(HomeroomPeriodsStep)


class FlourishFinalStep(flourish.content.ContentProvider, FinalStep):
    timetable = None

    getSessionData = lambda self: self.view.state
    next = lambda self: None

    def render(self, *args, **kw):
        # XXX: risky:
        if self.timetable is None:
            self.timetable = self.createAndAdd()
        self.request.response.redirect(
            absoluteURL(self.timetable, self.request))
        return ''



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2003 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Term implementation
"""
import persistent
import pytz
from datetime import datetime

import zope.interface
from zope.event import notify
from zope.proxy import sameProxiedObjects
from zope.component import adapts
from zope.component import adapter
from zope.interface import implements
from zope.interface import implementer
from zope.lifecycleevent.interfaces import IObjectRemovedEvent
from zope.container import contained, btree

from schooltool.schoolyear.subscriber import EventAdapterSubscriber
from schooltool.schoolyear.subscriber import ObjectEventAdapterSubscriber
from schooltool.schoolyear.interfaces import TermOverlapError
from schooltool.schoolyear.interfaces import ISubscriber
from schooltool.schoolyear.interfaces import ISchoolYear
from schooltool.schoolyear.interfaces import ISchoolYearContainer
from schooltool.app.interfaces import IApplicationPreferences
from schooltool.app.interfaces import ISchoolToolApplication
from schooltool.common import IDateRange
from schooltool.common import DateRange

from schooltool.term.interfaces import TermDateNotInSchoolYear
from schooltool.term import interfaces
from schooltool.common import SchoolToolMessage as _


class TermBeforeChangeEvent(object):

    def __init__(self, term, old_dates, new_dates):
        self.term = term
        self.old_dates = old_dates
        self.new_dates = new_dates


class TermAfterChangeEvent(object):

    def __init__(self, term, old_dates, new_dates):
        self.term = term
        self.old_dates = old_dates
        self.new_dates = new_dates


class EmergencyDayEvent(object):

    def __init__(self, date, replacement_date):
        self.date = date
        self.replacement_date = replacement_date


class Term(DateRange, contained.Contained, persistent.Persistent):
    zope.interface.implements(interfaces.ITerm, interfaces.ITermWrite)

    def __init__(self, title, first, last):
        self.title = title
        self._first = first
        self._last = last
        self._schooldays = set()
        if last < first:
            raise ValueError("Last date %r less than first date %r" %
                             (last, first))

    @property
    def first(self):
        return self._first

    @first.setter
    def first(self, new_first_date):
        old_dates = (self._first, self._last)
        new_dates = (new_first_date, self._last)

        if self._last < new_first_date:
            raise ValueError("Last date %r less than first date %r" %
                             (self._last, new_first_date))

        notify(TermBeforeChangeEvent(self, old_dates, new_dates))
        self._first = new_first_date
        notify(TermAfterChangeEvent(self, old_dates, new_dates))

    @property
    def last(self):
        return self._last

    @last.setter
    def last(self, new_last_date):
        old_dates = (self._first, self._last)
        new_dates = (self._first, new_last_date)

        if new_last_date < self._first:
            raise ValueError("Last date %r less than first date %r" %
                             (new_last_date, self._first))

        notify(TermBeforeChangeEvent(self, old_dates, new_dates))
        self._last = new_last_date
        notify(TermAfterChangeEvent(self, old_dates, new_dates))

    def _validate(self, date):
        if not date in self:
            raise ValueError("Date %r not in term [%r, %r]" %
                             (date, self.first, self.last))

    def isSchoolday(self, date):
        self._validate(date)
        if date in self._schooldays:
            return True
        return False

    def add(self, date):
        self._validate(date)
        self._schooldays.add(date)
        self._schooldays = self._schooldays  # persistence

    def remove(self, date):
        self._validate(date)
        self._schooldays.remove(date)
        self._schooldays = self._schooldays  # persistence

    def addWeekdays(self, *weekdays):
        for date in self:
            if date.weekday() in weekdays:
                self.add(date)

    def removeWeekdays(self, *weekdays):
        for date in self:
            if date.weekday() in weekdays and self.isSchoolday(date):
                self.remove(date)

    def toggleWeekdays(self, *weekdays):
        for date in self:
            if date.weekday() in weekdays:
                if self.isSchoolday(date):
                    self.remove(date)
                else:
                    self.add(date)

    def reset(self, first, last):
        if last < first:
            # import timemachine
            raise ValueError("Last date %r less than first date %r" %
                             (last, first))
        self.first = first
        self.last = last
        self._schooldays.clear()


class TermContainer(btree.BTreeContainer):
    """BBB: only there for backwards compatibility."""


def getTermForDate(date):
    """Find the term that contains `date`.

    Returns None if `date` falls outside all terms.
    """
    terms = interfaces.ITermContainer(date, {})
    for term in terms.values():
        if date in term:
            return term
    else:
        return None


def getNextTermForDate(date):
    """Find the term that contains `date`, or the next one.

    If there is a term that contains `date`, it is returned.  Otherwise, the
    first term that starts after `date` is returned.  If there are none,
    the last term that ended before `date` is returned.

    Returns None if there are no terms.
    """
    terms = interfaces.ITermContainer(date, {})
    before, after = [], []
    for term in terms.values():
        if date in term:
            return term
        if date > term.last:
            before.append((term.last, term))
        if date < term.first:
            after.append((term.first, term))
    if after:
        return min(after)[1]
    if before:
        return max(before)[1]
    return None


def listTerms(context):
    """List terms of a schoolyear in a chronological order."""
    terms = interfaces.ITermContainer(context, {})
    return sorted(terms.values(), key=lambda t: t.first)


def getPreviousTerm(term):
    """Get the next term in the same SchoolYear.

    The limitation is imposed as SchoolYears are strictly separated
    entities (for archival purposes).
    """
    prev_terms = [t for t in listTerms(term)
                  if t.last < term.first]
    if prev_terms:
        return prev_terms[-1]
    return None


def getNextTerm(term):
    """Get the next term in the same SchoolYear.

    The limitation is imposed as SchoolYears are strictly separated
    entities (for archival purposes).
    """
    next_terms = [t for t in listTerms(term)
                  if t.first > term.last]
    if next_terms:
        return next_terms[0]
    return None


class DateManagerUtility(object):
    zope.interface.implements(interfaces.IDateManager)

    @property
    def today(self):
        app = ISchoolToolApplication(None)
        tzinfo = pytz.timezone(IApplicationPreferences(app).timezone)
        dt = pytz.utc.localize(datetime.utcnow())
        return dt.astimezone(tzinfo).date()

    @property
    def current_term(self):
        return getNextTermForDate(self.today)


@implementer(interfaces.ITermContainer)
def getTermContainer(context):
    app = ISchoolToolApplication(None)
    syc = ISchoolYearContainer(app)
    return syc.getActiveSchoolYear()


@adapter(interfaces.ITerm)
@implementer(ISchoolYear)
def getSchoolYearForTerm(term):
    return term.__parent__


class RemoveTermsWhenSchoolYearIsDeleted(ObjectEventAdapterSubscriber):
    adapts(IObjectRemovedEvent, ISchoolYear)

    def __call__(self):
        for term_id in list(self.object.keys()):
            del self.object[term_id]


def validateTermsForOverlap(sy, dr, term):
    overlapping_terms = []
    for other_term in sy.values():
        if (not sameProxiedObjects(other_term, term) and
            dr.overlaps(IDateRange(other_term))):
            overlapping_terms.append(other_term)
            if overlapping_terms:
                raise TermOverlapError(term,
                                       overlapping_terms)


class TermOverlapValidationSubscriber(EventAdapterSubscriber):
    adapts(TermBeforeChangeEvent)
    implements(ISubscriber)

    def __call__(self):
        sy = self.event.term.__parent__
        dr = DateRange(*self.event.new_dates)
        if sy:
            validateTermsForOverlap(sy, dr, self.event.term)


class TermOverflowValidationSubscriber(EventAdapterSubscriber):
    adapts(TermBeforeChangeEvent)
    implements(ISubscriber)

    def __call__(self):
        if self.event.term.__parent__ is None:
            return
        dr = IDateRange(self.event.term.__parent__)
        if (self.event.new_dates[0] not in dr or
            self.event.new_dates[1] not in dr):
            raise TermDateNotInSchoolYear(_("Term date is not in the school year."))



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2005 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Setup code for SchoolTool application browser unit tests
"""
import os.path
import unittest

import transaction
from zope.interface import implements, Interface
from zope.component import provideAdapter, provideUtility
from zope.component.hooks import setSite
from zope.app.testing.functional import FunctionalTestSetup
from zope.app.testing import setup, ztapi
from zope.app.form.interfaces import IInputWidget
from zope.publisher.interfaces.browser import IBrowserRequest
from zope.publisher.interfaces.browser import IDefaultBrowserLayer
from zope.schema.interfaces import \
        IPassword, ITextLine, IText, IBytes, IBool, ISet, IList, IDate, \
        IInt, IChoice, IIterableVocabulary, IVocabularyTokenized, \
        ICollection
from zope.app.form.browser import \
        PasswordWidget, TextWidget, BytesWidget, CheckBoxWidget, \
        DateWidget, IntWidget, ChoiceInputWidget, DropdownWidget, \
        TextAreaWidget, ChoiceCollectionInputWidget, \
        CollectionInputWidget, MultiSelectWidget, OrderedMultiSelectWidget
from zope.app.form.interfaces import IWidgetInputError
from zope.app.form.browser.interfaces import IWidgetInputErrorView
from zope.app.form.browser.exception import WidgetInputErrorView
from zope.traversing.interfaces import ITraversable
from zope.traversing.namespace import view, resource
from zope.traversing.interfaces import IPathAdapter
from zope.browserpage.simpleviewclass import SimpleViewClass
from zope.app.basicskin.standardmacros import StandardMacros
from zope.app.form.browser.macros import FormMacros
from zope.browsermenu.menu import MenuAccessView
from zope.browsermenu.interfaces import IBrowserMenu
from zope.publisher.interfaces.browser import IBrowserView

from schooltool.relationship.tests import setUpRelationships
from schooltool.skin.macros import SchoolToolMacros

from schooltool.app.browser import SchoolToolAPI, SortBy

class BrowserMenuStub(object):
    """A stub that fakes browser menu.

    So we could display schooltool_actions menu in unit tests.
    """

    implements(IBrowserMenu)

    id = None
    title = None
    description = None

    def __init__(self, id, title=u'', description=u''):
        self.id = id
        self.title = title
        self.description = description

    def getMenuItems(self, object, request):
        return []


def setUp(test=None):
    """Set up the test fixture for schooltool.app.browser doctests.

    Performs what is called a "placeless setup" in the Zope 3 world, then sets
    up annotations, relationships, and registers widgets as views for some
    schema fields.

    In effect, duplicates a subset of ZCML configuration -- just enough to
    actually render our page templates in unit tests.
    """
    setup.placefulSetUp()
    setup.setUpAnnotations()
    setup.setUpTraversal()
    # relationships
    setUpRelationships()
    # widgets
    ztapi.browserViewProviding(IPassword, PasswordWidget, IInputWidget)
    ztapi.browserViewProviding(ITextLine, TextWidget, IInputWidget)
    ztapi.browserViewProviding(IText, TextAreaWidget, IInputWidget)
    ztapi.browserViewProviding(IBytes, BytesWidget, IInputWidget)
    ztapi.browserViewProviding(IBool, CheckBoxWidget, IInputWidget)
    ztapi.browserViewProviding(IDate, DateWidget, IInputWidget)
    ztapi.browserViewProviding(IInt, IntWidget, IInputWidget)
    ztapi.browserViewProviding(IChoice, ChoiceInputWidget, IInputWidget)
    ztapi.browserViewProviding(ICollection, CollectionInputWidget, IInputWidget)

    ztapi.provideMultiView((IChoice, IIterableVocabulary), IBrowserRequest,
                           IInputWidget, '', DropdownWidget)

    ztapi.provideMultiView((ISet, IChoice), IBrowserRequest,
                           IInputWidget, '', ChoiceCollectionInputWidget)
    ztapi.provideMultiView((IList, IChoice), IBrowserRequest,
                           IInputWidget, '', ChoiceCollectionInputWidget)
    ztapi.provideMultiView((IList, IVocabularyTokenized), IBrowserRequest,
                           IInputWidget, '', OrderedMultiSelectWidget)
    # XXX MultiSelectWidget doesn't work with sets :/
    #     http://www.zope.org/Collectors/Zope3-dev/360
    ztapi.provideMultiView((ISet, IIterableVocabulary), IBrowserRequest,
                           IInputWidget, '', MultiSelectWidget)

    # errors in forms
    ztapi.browserViewProviding(IWidgetInputError, WidgetInputErrorView,
                               IWidgetInputErrorView)


    # Now, the question is: does the speed of the tests run with the
    # setup below justify this complex setup that duplicates the ZCML?
    # For now, I say yes. -- not mg, perhaps alga or gintas

    # ++view++
    ztapi.provideView(None, None, ITraversable, 'view', view)
    ztapi.provideView(None, None, ITraversable, 'resource', resource)

    # schooltool: namespace in tal
    provideAdapter(SchoolToolAPI, (None,), IPathAdapter, 'schooltool')

    # sortby: namespace in tal
    provideAdapter(SortBy, (None,), IPathAdapter, 'sortby')

    # standard_macros, schooltool_macros and schooltool_navigation
    ztapi.browserView(None, 'standard_macros', StandardMacros)
    ztapi.browserView(None, 'view_macros',
                      SimpleViewClass("../../skin/templates/view_macros.pt"))

    ztapi.browserView(None, 'schooltool_macros', SchoolToolMacros)
    ztapi.browserView(None, 'calendar_macros',
                      SimpleViewClass("./templates/calendar_macros.pt"))
    ztapi.browserView(None, 'generic_macros',
                      SimpleViewClass("../../skin/templates/generic_macros.pt"))

    # form macros
    ztapi.browserView(None, 'form_macros', FormMacros)
    import zope.formlib
    base = zope.formlib.__path__[0]
    ztapi.browserView(None, 'widget_macros',
                      SimpleViewClass(os.path.join(base, 'widget_macros.pt')))

    # resources
    class ResourceStub:
        def __init__(self, request):
            self.request = request
        def __getitem__(self, key):
            return ResourceStub(self.request)
        def __call__(self):
            return "a dummy resource"

    for name in ['layout.css', 'schooltool.css', 'schooltool.js',
                 'next.png', 'prev.png', 'favicon.ico',
                 'print.css', 'jquery.js',
                 'zonki-regular.png']:
        ztapi.browserResource(name, ResourceStub)

    # menus
    ztapi.browserView(None, 'view_get_menu', MenuAccessView)
    provideUtility(BrowserMenuStub('schooltool_actions'), IBrowserMenu,
                   'schooltool_actions')

    # `provider` TALES namespaces
    from zope.browserpage import metaconfigure
    from zope.contentprovider import tales
    metaconfigure.registerType('provider', tales.TALESProviderExpression)

    # viewlet manager registrations
    from zope.viewlet import manager
    from schooltool import skin
    name = 'schooltool.Header'
    provideAdapter(
        manager.ViewletManager(name, skin.IHeaderManager),
        (Interface, IDefaultBrowserLayer, IBrowserView),
        skin.IHeaderManager,
        name=name)

    name = 'schooltool.JavaScript'
    provideAdapter(
        manager.ViewletManager(name, skin.IJavaScriptManager),
        (Interface, IDefaultBrowserLayer, IBrowserView),
        skin.IJavaScriptManager,
        name=name)

    name = 'schooltool.CSS'
    provideAdapter(
        manager.ViewletManager(name, skin.ICSSManager),
        (Interface, IDefaultBrowserLayer, IBrowserView),
        skin.ICSSManager,
        name=name)

    name = 'schooltool.MenuBar'
    provideAdapter(
        manager.ViewletManager(name, skin.skin.IMenuBarMenuManager),
        (Interface, IDefaultBrowserLayer, IBrowserView),
        skin.skin.IMenuBarMenuManager,
        name=name)

    name = 'schooltool.NavigationMenu'
    provideAdapter(
        manager.ViewletManager(name, skin.skin.INavigationManager),
        (Interface, IDefaultBrowserLayer, IBrowserView),
        skin.skin.INavigationManager,
        name=name)

    name = 'schooltool.ActionsMenu'
    provideAdapter(
        manager.ViewletManager(name, skin.skin.IActionMenuManager),
        (Interface, IDefaultBrowserLayer, IBrowserView),
        skin.skin.IActionMenuManager,
        name=name)


def tearDown(test=None):
    """Tear down the test fixture for schooltool.app.browser doctests."""
    transaction.abort()
    setup.placefulTearDown()


def layeredTestSetup():
    fts = FunctionalTestSetup()
    fts.setUp()
    app = fts.getRootFolder()
    setSite(app)


def layeredTestTearDown():
    setSite(None)
    fts = FunctionalTestSetup()
    fts.tearDown()


def makeLayeredSuite(klass, layer):
    suite = unittest.makeSuite(klass)
    suite.layer = layer
    return suite



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2006 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Unit tests for schooltool.securitypolicy.crowds
"""
import unittest
import doctest

from zope.app.testing import setup
from zope.interface import implements
from zope.interface.verify import verifyObject
from zope.component import queryUtility

from schooltool.securitypolicy.interfaces import IAccessControlCustomisations
from schooltool.app.interfaces import ISchoolToolApplication
from schooltool.securitypolicy.crowds import CrowdsUtility, DescriptionUtility
from schooltool.securitypolicy.interfaces import ICrowdsUtility
from schooltool.securitypolicy.interfaces import IDescriptionUtility


def doctest_CrowdsUtility():
    """Doctest for CrowdsUtility.

        >>> cru = CrowdsUtility()
        >>> verifyObject(ICrowdsUtility, cru)
        True
    """


def doctest_getCrowdsUtility():
    """Doctest for getCrowdsUtility.

        >>> from schooltool.securitypolicy.crowds import getCrowdsUtility
        >>> queryUtility(ICrowdsUtility) is None
        True

        >>> cru = getCrowdsUtility()
        >>> print cru
        <schooltool.securitypolicy.crowds.CrowdsUtility object ...>

        >>> queryUtility(ICrowdsUtility) is cru
        True

    """


def doctest_DescriptionUtility():
    """Doctest for DescriptionUtility.

        >>> du = DescriptionUtility()
        >>> verifyObject(IDescriptionUtility, du)
        True

    """


def doctest_getDescriptionUtility():
    """Doctest for getDescriptionUtility.

        >>> from schooltool.securitypolicy.crowds import getDescriptionUtility
        >>> queryUtility(IDescriptionUtility) is None
        True

        >>> du = getDescriptionUtility()
        >>> print du
        <schooltool.securitypolicy.crowds.DescriptionUtility object ...>

        >>> queryUtility(IDescriptionUtility) is du
        True

    """


def doctest_AggregateCrowd():
    """Doctests for AggregateCrowd.

    What's so special about this adapter?  It aggregates crowds retrieved from
    objcrowds:

        >>> class CrowdStub(object):
        ...     def __init__(self, context):
        ...         pass
        ...     def contains(self, principal):
        ...         print 'contains(%s)' % principal
        ...         return principal == 'r00t'

    The crowdFactories method is normally overridden:

        >>> from schooltool.securitypolicy.crowds import AggregateCrowd
        >>> adapter = AggregateCrowd(object())

        >>> adapter.crowdFactories = lambda: [CrowdStub]

        >>> adapter.contains('some principal')
        contains(some principal)
        False

        >>> adapter.contains('r00t')
        contains(r00t)
        True

    """


def doctest_ConfigurableCrowd():
    """Tests for ConfigurableCrowd.

    Some setup:

        >>> setup.placelessSetUp()
        >>> class CustomisationsStub(object):
        ...     implements(IAccessControlCustomisations)
        ...     def get(self, key):
        ...         print 'Getting %s' % key
        ...         return True

        >>> class AppStub(object):
        ...     implements(ISchoolToolApplication)
        ...     def __conform__(self, iface):
        ...         if iface == IAccessControlCustomisations:
        ...             return CustomisationsStub()

        >>> from zope.component import provideAdapter
        >>> provideAdapter(lambda context: AppStub(),
        ...                adapts=[None],
        ...                provides=ISchoolToolApplication)

        >>> from schooltool.securitypolicy.crowds import ConfigurableCrowd

    Off we go:

        >>> crowd = ConfigurableCrowd(object())
        >>> crowd.setting_key = 'key'
        >>> crowd.contains(object())
        Getting key
        True

    Clean up:

        >>> setup.placelessTearDown()

    """


def doctest_ParentCrowd():
    """Tests for ParentCrowd.

        >>> setup.placelessSetUp()

        >>> from schooltool.securitypolicy.interfaces import ICrowd
        >>> class ICalendarParentCrowd(ICrowd):
        ...     pass
        >>> class CalendarStub(object):
        ...     def __init__(self, parent):
        ...         self.__parent__ = parent

        >>> from schooltool.securitypolicy.crowds import ParentCrowd
        >>> CalendarViewersCrowd = ParentCrowd(ICalendarParentCrowd,
        ...                                    'schooltool.view')
        >>> crowd = CalendarViewersCrowd(CalendarStub(None))
        >>> crowd.perm = "schooltool.view"

    First, fire a blank (no adapters registered):

        >>> crowd.contains(object())
        False

    OK, let's try with an adaptable object now:

        >>> from schooltool.app.interfaces import ICalendarParentCrowd

        >>> class ParentCrowdStub(object):
        ...     implements(ICalendarParentCrowd)
        ...     def __init__(self, context):
        ...         print 'Getting adapter for %s' % context
        ...     def contains(self, principal):
        ...         print 'Checking %s' % principal
        ...         return True

        >>> from zope.interface import Interface
        >>> class IOwner(Interface):
        ...     pass

        >>> class OwnerStub(object):
        ...     implements(IOwner)

        >>> from zope.component import provideAdapter
        >>> provideAdapter(ParentCrowdStub, adapts=[IOwner],
        ...                provides=ICalendarParentCrowd,
        ...                name='schooltool.view')

    Let's try now with the adapter:

        >>> from schooltool.app.security import CalendarViewersCrowd
        >>> crowd = CalendarViewersCrowd(CalendarStub(OwnerStub()))
        >>> crowd.contains('some principal')
        Getting adapter for <...OwnerStub ...>
        Checking some principal
        True

    Let's try another name:

        >>> CalendarEditorsCrowd = ParentCrowd(ICalendarParentCrowd,
        ...                                    'schooltool.edit')
        >>> crowd = CalendarEditorsCrowd(CalendarStub(OwnerStub()))
        >>> crowd.contains('some principal')
        False

    Now with an adapter in place:

        >>> from zope.component import provideAdapter
        >>> provideAdapter(ParentCrowdStub, adapts=[IOwner],
        ...                provides=ICalendarParentCrowd,
        ...                name='schooltool.edit')
        >>> crowd = CalendarEditorsCrowd(CalendarStub(OwnerStub()))
        >>> crowd.contains('some principal')
        Getting adapter for <...OwnerStub ...>
        Checking some principal
        True

    We're done.

        >>> setup.placefulTearDown()

    """


def test_suite():
    return unittest.TestSuite([
            doctest.DocTestSuite(optionflags=doctest.ELLIPSIS |
                                             doctest.NORMALIZE_WHITESPACE)])



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2005 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Person sample data generation
"""

from zope.interface import implements

from schooltool.sampledata.interfaces import ISampleDataPlugin
from schooltool.resource.resource import Resource, Location


class SampleResources(object):

    implements(ISampleDataPlugin)

    name = 'resources'
    dependencies = ()

    def generate(self, app, seed=None):
        for i in range(64):
            room = app['resources']['room%02d' % i] = Location(title='Room %02d' % i)
            room.type = "Classroom"
        for i in range(24):
            resource = Resource(title='Projector %02d' % i)
            projector = app['resources']['projector%02d' % i] = resource
            projector.type = "Projector"



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2005 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Unit tests for schooltool.sampledata.generator
"""

import unittest
import doctest

from zope.interface import implements
from zope.app.testing import setup
from zope.component import provideUtility

from schooltool.sampledata.interfaces  import ISampleDataPlugin


class DummyPlugin:
    """A dummy sample data plugin that logs the args to a class attribute"""
    implements(ISampleDataPlugin)

    log = []  # one for all

    def __init__(self, name, deps=()):
        self.name = name
        self.dependencies = deps

    def generate(self, app, seed=None):
        self.log.append((self.name, app, seed))


def doctest_generate():
    """Unit test for schooltool.sampledata.generator.generate

    This function finds all the utilities registered with the sample
    data plugin interface and runs the generate methods of them.

    First, let's register several plugins:

        >>> p1 = DummyPlugin("p1", ())
        >>> p2 = DummyPlugin("p2", ())
        >>> p3 = DummyPlugin("p3", ())
        >>> provideUtility(p1, ISampleDataPlugin, 'p1')
        >>> provideUtility(p2, ISampleDataPlugin, 'p2')
        >>> provideUtility(p3, ISampleDataPlugin, 'p3')

    Now, let's run the generator:

        >>> DummyPlugin.log = []
        >>> import schooltool.sampledata.generator

        >>> app = 'app'
        >>> result = schooltool.sampledata.generator.generate(app,
        ...                 pluginNames=['p1','p2','p3'])

    The order of the plugins is undefined, so let's sort to see if
    they're all there.

        >>> DummyPlugin.log.sort()
        >>> DummyPlugin.log
        [('p1', 'app', None), ('p2', 'app', None), ('p3', 'app', None)]

    The result returned by this function is a dict of plugin names as
    keys and CPU times consumed as values:

        >>> sorted(result.keys())
        ['p1', 'p2', 'p3']
        >>> for v in result.values():
        ...     v < 0.1
        True
        True
        True

    Let's make p1 and p2 depend on p3:

        >>> p1.dependencies = 'p3',
        >>> p2.dependencies = 'p3',

    Now p3 must be the first plugin run:

        >>> DummyPlugin.log = []
        >>> result = schooltool.sampledata.generator.generate(app,
        ...                 pluginNames=['p1','p2','p3'])
        >>> DummyPlugin.log[0]
        ('p3', 'app', None)

    If we add a dependency on p2 to p1, the order becomes defined:

        >>> p1.dependencies = 'p2',
        >>> DummyPlugin.log = []
        >>> result = schooltool.sampledata.generator.generate(app,
        ...                 pluginNames=['p1','p2','p3'])
        >>> DummyPlugin.log
        [('p3', 'app', None), ('p2', 'app', None), ('p1', 'app', None)]

    If we add an extra dependency and close the cycle, we get an error:

        >>> p3.dependencies = 'p1',
        >>> DummyPlugin.log = []
        >>> result = schooltool.sampledata.generator.generate(app,
        ...                 pluginNames=['p1','p2','p3'])
        Traceback (most recent call last):
          ...
        CyclicDependencyError: cyclic dependency at ...

    """


def test_suite():
    return unittest.TestSuite([
        doctest.DocTestSuite(setUp=setup.placelessSetUp,
                             tearDown=setup.placelessTearDown,
                             optionflags=doctest.ELLIPSIS
                             |doctest.NORMALIZE_WHITESPACE),
        ])

if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2009 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Inline page templates
"""

from zope.pagetemplate.engine import TrustedAppPT
from zope.pagetemplate.pagetemplate import PageTemplate
from zope.browserpage.viewpagetemplatefile import BoundPageTemplate


class InlineTemplateBase(TrustedAppPT, PageTemplate):

    def __init__(self, source, content_type=None):
        self.source = source
        if content_type is not None:
            self.content_type = content_type
        self.pt_edit(source, self.content_type)


class InlinePageTemplate(InlineTemplateBase):
    """Inline page template.

    Use it like this:

        >>> pt = InlinePageTemplate('''
        ... <ul>
        ...   <li tal:repeat="item items" tal:content="item">(item)</li>
        ... </ul>
        ... ''')
        >>> print pt(items=['a', 'b', 'c']).strip()
        <ul>
          <li>a</li>
          <li>b</li>
          <li>c</li>
        </ul>

    """

    def pt_getContext(self, args, options):
        namespace = {'template': self, 'args': args, 'nothing': None}
        namespace.update(self.pt_getEngine().getBaseNames())
        namespace.update(options)
        return namespace


class InlineViewPageTemplate(InlineTemplateBase):
    """Inline page template for views.

    Use it like this:

        >>> class View(object):
        ...     template = InlineViewPageTemplate('''
        ...         <p tal:content="context"></p>
        ...         <p>Page <tal:c content="request/page"></tal:c></p>
        ...         <p tal:content="view"></p>
        ...         <ul>
        ...           <li tal:repeat="item options/items"
        ...               tal:content="item">
        ...           </li>
        ...         </ul>
        ...         ''')
        ...     def __init__(self, context, request):
        ...         self.context = context
        ...         self.request = request
        ...     __repr__ = lambda self: '%s object' % self.__class__.__name__

        >>> from zope.publisher.browser import TestRequest
        >>> view = View('the context', TestRequest(form={'page': 5}))
        >>> print view.template(items=['a', 'b']).strip()
        <p>the context</p>
        <p>Page 5</p>
        <p>View object</p>
        <ul>
          <li>a</li>
          <li>b</li>
        </ul>

    """

    def pt_getContext(self, instance, **kw):
        namespace = InlineTemplateBase.pt_getContext(self, **kw)
        namespace.update({
                'context': instance.context,
                'request': instance.request,
                'view': instance,
                })
        return namespace

    def __call__(self, instance, *args, **keywords):
        namespace = self.pt_getContext(
            instance, args=args, options=keywords)
        s = self.pt_render(
            namespace,
            showtal=getattr(instance.request.debug, 'showTAL', False),
            )
        response = instance.request.response
        if not response.getHeader("Content-Type"):
            response.setHeader("Content-Type", self.content_type)
        return s

    def __get__(self, instance, type):
        return BoundPageTemplate(self, instance)


def InheritTemplate(template):
    """Re-bind templates from other classes.

        >>> class Foo(object):
        ...     template = InlineViewPageTemplate("")
        ...     other_template = InlinePageTemplate("")

        >>> class Bar(object):
        ...     template = InheritTemplate(Foo.template)
        ...     other_template = InheritTemplate(Foo.other_template)

        >>> Foo().template
        <BoundPageTemplateFile of <schooltool.common.inlinept.Foo ...>>

        >>> Bar().template
        <BoundPageTemplateFile of <schooltool.common.inlinept.Bar ...>>

    Unbound templates are returned unaffected.

        >>> Bar().other_template
        <schooltool.common.inlinept.InlinePageTemplate ...>

    """
    if isinstance(template, BoundPageTemplate):
        return template.im_func
    return template



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2005 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
csv importing.
"""
from zope.component import getUtility
from zope.container.interfaces import INameChooser

from schooltool.app.browser.csvimport import BaseCSVImporter, BaseCSVImportView
from schooltool.person.interfaces import IPersonFactory

from schooltool.common import SchoolToolMessage as _

class PersonCSVImporter(BaseCSVImporter):
    """A Person CSV importer."""

    def createAndAdd(self, data, dry_run=True):
        """Create Person object and add to container.

        We are requiring that we have a username and fullname (title) set.  If
        any duplicates are found then an error is reported and the duplicate
        entries are reported back to the user.
        """
        if len(data) < 2:
            self.errors.fields.append(_("""Insufficient data provided."""))
            return

        if not data[0]:
            self.errors.fields.append(_('username may not be empty'))
            return

        if not data[1]:
            self.errors.fields.append(_('fullname may not be empty'))
            return

        username = data[0]
        fullname = data[1]
        if len(data) > 2:
            password = data[2]
        else:
            password = None

        if username in self.container:
            error_msg = _("Duplicate username: ${username}",
                          mapping={'username' : ', '.join(data)})
            self.errors.fields.append(error_msg)
            return

        try:
            INameChooser(self.container).checkName(username, None)
        except ValueError:
            error_msg = _("Names cannot begin with '+' or '@' or contain '/'")
            self.errors.fields.append(error_msg)
            return

        obj = self.personFactory(data[0], data[1])

        if password:
            obj.setPassword(password)

        if not dry_run:
            self.container[data[0]] = obj

    def personFactory(self, username, title):
        factory = getUtility(IPersonFactory)
        person = factory(username=username, title=title)
        return person

class PersonCSVImportView(BaseCSVImportView):
    """View for Person CSV importer."""

    importer_class = PersonCSVImporter



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2010 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
SchoolTool catalogs.
"""
from zope.interface import implementer, implements, implementsOnly
from zope.intid.interfaces import IIntIds, IIntIdAddedEvent, IIntIdRemovedEvent
from zope.component import adapter, queryUtility, getUtility
from zope.container import btree
from zope.container.contained import Contained
from zope.lifecycleevent import IObjectModifiedEvent
from zope.security.proxy import removeSecurityProxy

from zc.catalog import extentcatalog
from zc.catalog import catalogindex

from schooltool.app.interfaces import ISchoolToolApplication
from schooltool.app.interfaces import ICatalogStartUp
from schooltool.app.interfaces import ICatalogs
from schooltool.app.interfaces import IVersionedCatalog
from schooltool.app.app import ActionBase
from schooltool.table.catalog import FilterImplementing


APP_CATALOGS_KEY = 'schooltool.app.catalog:Catalogs'


class CatalogStartupBase(ActionBase):
    implementsOnly(ICatalogStartUp)


class Catalogs(btree.BTreeContainer):
    implements(ICatalogs)


@adapter(ISchoolToolApplication)
@implementer(ICatalogs)
def getAppCatalogs(app):
    if APP_CATALOGS_KEY not in app:
        app[APP_CATALOGS_KEY] = Catalogs()
    return app[APP_CATALOGS_KEY]


class VersionedCatalog(Contained):
    implements(IVersionedCatalog)

    expired = False
    version = 0
    catalog = None

    def __init__(self, catalog, version):
        self.catalog = catalog
        self.catalog.__parent__ = self
        self.catalog.__name__ = 'catalog'
        self.version = version

    def __repr__(self):
        return '<%s v. %r>: %s' % (
            self.__class__.__name__, self.version, self.catalog)


class PrepareCatalogContainer(CatalogStartupBase):

    def __call__(self):
        catalogs = ICatalogs(self.app)
        for entry in catalogs.values():
            entry.expired = True


class ExpiredCatalogCleanup(CatalogStartupBase):

    def __call__(self):
        catalogs = ICatalogs(self.app)
        for key in list(catalogs):
            if catalogs[key].expired:
                del catalogs[key]


class CatalogFactory(CatalogStartupBase):

    after = ('prepare-catalog-container', )
    before = ('expired-catalog-cleanup', )

    version = u''

    @classmethod
    def key(cls):
        return u'catalog:%s.%s' % (cls.__module__, cls.__name__)

    @classmethod
    def get(cls, ignored=None):
        app = ISchoolToolApplication(None)
        catalogs = ICatalogs(app)
        entry = catalogs.get(cls.key())
        if entry is None:
            return None
        return entry.catalog

    def getVersion(self):
        return unicode(self.version)

    def createCatalog(self):
        raise NotImplementedError()

    def setIndexes(self, catalog):
        raise NotImplementedError()

    def __call__(self):
        app = ISchoolToolApplication(None)
        catalogs = ICatalogs(app)
        key = self.key()
        version = self.getVersion()

        if key in catalogs:
            if catalogs[key].version == version:
                catalogs[key].expired = False
            else:
                del catalogs[key]

        if key not in catalogs:
            catalog = self.createCatalog()
            catalogs[key] = VersionedCatalog(catalog, version)
            # XXX: if setIndexes throw, delete the catalog and rethrow
            self.setIndexes(catalog)


class CatalogImplementing(CatalogFactory):
    """Factory of catalogs containing objects implementing the given interface."""

    interface = None # override in child classes

    def createCatalog(self):
        return extentcatalog.Catalog(
            extentcatalog.FilterExtent(
                FilterImplementing(self.interface)))

    def getVersion(self):
        return u'interface:%s, version:%s' % (
            u'%s.%s' % (self.interface.__module__, self.interface.__name__),
            super(CatalogImplementing, self).getVersion())


class AttributeCatalog(CatalogImplementing):
    """Catalog indexing specified attributes of objects implementing
    the given interface."""

    attributes = ()

    def getVersion(self):
        return u'attributes:%s, %s' % (
            tuple(sorted(self.attributes)),
            super(AttributeCatalog, self).getVersion())

    def setIndexes(self, catalog):
        for name in self.attributes:
            catalog[name] = catalogindex.ValueIndex(name)


@adapter(IIntIdAddedEvent)
def indexDocSubscriber(event):
    app = ISchoolToolApplication(None, None)
    if app is None:
        return
    obj = removeSecurityProxy(event.object)
    util = getUtility(IIntIds, context=app)
    obj_id = util.getId(obj)
    catalogs = ICatalogs(app)
    for entry in catalogs.values():
        entry.catalog.index_doc(obj_id, obj)


@adapter(IObjectModifiedEvent)
def reindexDocSubscriber(event):
    app = ISchoolToolApplication(None, None)
    if app is None:
        return
    obj = removeSecurityProxy(event.object)
    util = queryUtility(IIntIds, context=app)
    if util is None:
        return
    obj_id = util.queryId(obj)
    if obj_id is None:
        return
    catalogs = ICatalogs(app)
    if obj is catalogs:
        return
    for entry in catalogs.values():
        entry.catalog.index_doc(obj_id, obj)


@adapter(IIntIdRemovedEvent)
def unindexDocSubscriber(event):
    app = ISchoolToolApplication(None, None)
    if app is None:
        return
    obj = removeSecurityProxy(event.object)
    util = getUtility(IIntIds, context=app)
    obj_id = util.queryId(obj)
    if obj_id is None:
        return
    catalogs = ICatalogs(app)
    for entry in catalogs.values():
        entry.catalog.unindex_doc(obj_id)


def appendGlobbing(text):
    words = filter(None, text.split(' '))
    return ' '.join([word.endswith('*') and word or ('%s*' % word)
                     for word in words])


def buildQueryString(text):
    terms = []
    parts = text.lower().split('"')
    for i, part in enumerate(parts):
        if i % 2 == 0:
            # before quote
            if not part:
                continue
            alternatives = [term.strip()
                            for term in part.split(',')]
            alternatives = map(appendGlobbing, alternatives)
            alternatives = filter(None, alternatives)
            terms.append(' or '.join(alternatives))
        else:
            # insert text inside quotes verbatim
            terms.append('"%s"' % part)
    return ' '.join(terms)



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2006 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
SchoolTool security policy metadirectives.

"""

from zope.interface import Interface
from zope import schema
from zope.configuration import fields
from zope.security.zcml import Permission


CrowdId = fields.PythonIdentifier
DescGroupId = fields.PythonIdentifier
DescActionId = fields.PythonIdentifier


class ICrowdDirective(Interface):

    name = CrowdId(
        title=u"Name",
        description=u"Identifier of the crowd")

    factory = fields.GlobalObject(
        title=u"Factory",
        description=u"The factory for the crowd")


class IAllowDirective(Interface):

    interface = fields.GlobalObject(
        title=u"Interface",
        required=False,
        description=u"An interface of the object")

    crowds = fields.Tokens(
        title=u"Crowds",
        value_type=CrowdId(title=u"Crowd"),
        required=True)

    permission = Permission(
        title=u"Permission",
        required=True)


class ISettingDirective(Interface):

    key = fields.PythonIdentifier(
        title=u"Key")

    text = fields.MessageID(
        title=u"Text that describes the setting")

    alt_text = fields.MessageID(
        title=u"Text that describes 'setting=False'",
        required=False)

    default = fields.Bool(
        title=u"Default setting",
        required=True,
        default=False)


class IAggregateCrowdDirective(Interface):

    name = CrowdId(
        title=u"Name",
        description=u"Identifier of the crowd")

    crowds = fields.Tokens(
        title=u"Crowds",
        value_type=CrowdId(title=u"Crowd"),
        required=True)


class IDescribeGroup(Interface):

    name = DescGroupId(
        title=u"Name",
        description=u"Unique identifier of the group.")

    title = fields.MessageID(
        title=u"Title",
        description=u"Group title displayed to the user.",
        required=False)

    description = fields.MessageID(
        title=u"Description",
        description=u"Group description displayed to the user.",
        required=False)

    klass = fields.GlobalObject(
        title=u"Class",
        description=u"""
        Group definition class that builds it's own title/description.
        This is an alternative to title/description defined in ZCML.
        """,
        required=False)


class IDescribeAction(Interface):

    group = DescGroupId(
        title=u"Group name",
        description=u"Identifier of the group this action belongs to.")

    name = DescActionId(
        title=u"Name",
        description=u"Unique identifier of the action within the group.")

    order = schema.Int(
        title=u"Order",
        description=u"Order in which this action should be displayed.",
        required=False)

    interface = fields.GlobalObject(
        title=u"Interface",
        required=True)

    permission = Permission(
        title=u"Permission",
        required=True)

    title = fields.MessageID(
        title=u"Title",
        required=False)

    description = fields.MessageID(
        title=u"Description",
        required=False)

    klass = fields.GlobalObject(
        title=u"Class",
        description=u"""
        Action definition class that builds it's own title/description.
        This is an alternative to title/description defined in ZCML.
        """,
        required=False)


class IDescribeCrowd(Interface):

    group = DescGroupId(
        title=u"Group",
        description=u"""
        Optional identifier of the group.  When specified, this description applies
        to the group only.
        """,
        required=False)

    action = DescActionId(
        title=u"Action",
        description=u"""
        Optional identifier of the action of the group.  If specified, this description
        applies to the action only.
        """,
        required=False)

    crowd = CrowdId(
        title=u"Crowd",
        description=u"Identifier of the crowd.",
        required=False)

    crowd_factory = fields.GlobalObject(
        title=u"Crowd factory",
        description=u"Alternative way to specify the crowd.",
        required=False)

    factory = fields.GlobalObject(
        title=u"Description class",
        description=u"Optional class that to build the title/description.",
        required=False)

    title = fields.MessageID(
        title=u"Title",
        description=u"""
        A quick way to specify the title from ZCML.
        It will be assigned to the "factory" instance dict.
        """,
        required=False)

    description = fields.MessageID(
        title=u"Description",
        description=u"""
        A quick way to specify the description from ZCML.
        It will be assigned to the "factory" instance dict.
        """,
        required=False)


class ISwitchDescription(Interface):
    """Directive to use a replacement crowd when looking up descriptions of
    a crowd.  May be used to use the replacement for a specific group or action.
    """

    group = DescGroupId(
        title=u"Group",
        description=u"""
        Optional identifier of the group.  When specified,
        the description will be switched within this group only.
        """,
        required=False)

    action = DescActionId(
        title=u"Action",
        description=u"""
        Optional identifier of the action of the group.
        When specified, the description will be switched for
        this action of the group only.
        """,
        required=False)

    crowd = CrowdId(
        title=u"Crowd to replace",
        description=u"Identifier of the crowd to replace.",
        required=False)

    crowd_factory = fields.GlobalObject(
        title=u"Crowd to replace factory",
        description=u"Alternative way to specify the crowd to replace.",
        required=False)

    use_crowd = CrowdId(
        title=u"Crowd",
        description=u"Identifier of the crowd to use for description.",
        required=False)

    use_crowd_factory = fields.GlobalObject(
        title=u"Crowd factory",
        description=u"Alternative way to specify the description crowd",
        required=False)





#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2011 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
SchoolTool calendar event views.
"""
import urllib
import datetime

import zope.formlib.interfaces
import zope.formlib.widget
from zope import formlib
from zope.i18n import translate
from zope.interface import Interface
from zope.cachedescriptors.property import Lazy
from zope.component import getMultiAdapter
from zope.security.proxy import removeSecurityProxy
from zope.traversing.browser.absoluteurl import absoluteURL

import zc.table.table
import z3c.form.browser.radio
from z3c.form import form, field, button, widget, validator

import schooltool.skin.flourish.page
import schooltool.skin.flourish.form
from schooltool.app.browser import ViewPreferences
from schooltool.app.browser.cal import CalendarEventView
from schooltool.app.browser.cal import ICalendarEventAddForm
from schooltool.app.browser.cal import CalendarEventAddView
from schooltool.app.browser.cal import ICalendarEventEditForm
from schooltool.app.browser.cal import CalendarEventEditView
from schooltool.app.browser.cal import CalendarEventBookingView
from schooltool.app.utils import vocabulary
from schooltool.calendar.app import CalendarEvent
from schooltool.calendar.interfaces import ICalendar
from schooltool.calendar.utils import parse_date
from schooltool.skin import flourish
from schooltool.table.table import ImageInputColumn
from schooltool.table.table import label_cell_formatter_factory

from schooltool.common import SchoolToolMessage as _


calendarFieldNames = (
    "title", "allday", "start_date", "start_time", "duration",
    "duration_type", "recurrence", "location", "description",
    "recurrence_type", "interval", "range", "until", "count",
    "monthly", "exceptions", "weekdays",
    )


class FlourishCalendarEventView(flourish.page.Page,
                                CalendarEventView):
    pass


class FlourishCalendarEventEditView(flourish.page.Page,
                                    CalendarEventEditView):
    schema = ICalendarEventEditForm
    fieldNames = calendarFieldNames

    update = CalendarEventEditView.update

    def setCustomWidget(self, name, **kw):
        factory = lambda field, request: getMultiAdapter(
            (field, self.request), formlib.interfaces.IInputWidget)
        custom_factory = formlib.widget.CustomWidgetFactory(
            factory, cssClass="date-field", **kw)
        setattr(self, str(name+'_widget'), custom_factory)

    def _setUpWidgets(self):
        self.setCustomWidget('description', height=5)
        self.setCustomWidget('exceptions', width=20, height=5)
        super(FlourishCalendarEventEditView, self)._setUpWidgets()


class FlourishCalendarEventAddView(flourish.page.Page,
                                   CalendarEventAddView):
    schema = ICalendarEventAddForm
    fieldNames = calendarFieldNames

    update = CalendarEventAddView.update

    _arguments = ()
    _keyword_arguments = calendarFieldNames
    _set_after_add = ()
    _set_before_add = None
    _factory = CalendarEvent

    def setCustomWidget(self, name, **kw):
        factory = lambda field, request: getMultiAdapter(
            (field, self.request), formlib.interfaces.IInputWidget)
        custom_factory = formlib.widget.CustomWidgetFactory(factory, **kw)
        setattr(self, str(name+'_widget'), custom_factory)

    def setUpCustomWidgets(self):
        self.setCustomWidget('description', height=5)
        self.setCustomWidget('exceptions', width=20, height=5)

    def _setUpWidgets(self):
        self.setUpCustomWidgets()
        super(FlourishCalendarEventAddView, self)._setUpWidgets()


class FlourishCalendarEventBookingView(flourish.page.Page,
                                       CalendarEventBookingView):

    update = CalendarEventBookingView.update

    def renderBookedTable(self):
        prefix = "remove_item"
        available_columns = self.columns()
        available_columns[0].cell_formatter = label_cell_formatter_factory(prefix)
        available_columns[2].title = _('Reserved by others')
        columns = list(available_columns)

        title=_('Release resource')
        # XXX: this getter is just plain wrong
        getter = lambda r: r.__name__

        remove_column = ImageInputColumn(
            prefix, name='action',
            title=title, alt=title,
            library='schooltool.skin.flourish',
            image='remove-icon.png', id_getter=getter)
        columns.append(remove_column)

        formatter = zc.table.table.FormFullFormatter(
            self.context, self.request, self.getBookedItems(),
            columns=columns,
            sort_on=self.sortOn(),
            prefix="booked")
        formatter.cssClasses['table'] = 'data'

        return formatter()

    def renderAvailableTable(self):
        prefix = "add_item"
        available_columns = self.columns()
        available_columns[0].cell_formatter = label_cell_formatter_factory(prefix)
        available_columns[2].title = _('Reserved by others')
        columns = list(available_columns)

        title=_('Reserve resource')
        # XXX: this getter is just plain wrong
        getter = lambda r: r.__name__

        add_column = ImageInputColumn(
            prefix, name='action',
            title=title, alt=title,
            library='schooltool.skin.flourish',
            image='add-icon.png', id_getter=getter)
        columns.append(add_column)

        formatter = zc.table.table.FormFullFormatter(
            self.context, self.request, self.filter(self.availableResources),
            columns=columns,
            batch_start=self.batch.start, batch_size=self.batch.size,
            sort_on=self.sortOn(),
            prefix="available")
        formatter.cssClasses['table'] = 'data'


        return formatter()


class IDeleteRecurringEventForm(Interface):

    delete = zope.schema.Choice(
        title=_("Delete"),
        vocabulary=vocabulary([
                ("all", _("All occurrences of this event")),
                ("current", _("Only current occurrence")),
                ("future", _("This and all future occurrences"))]),
        default="current",
        required=True,
        )


class DeleteEventDialog(flourish.form.DialogForm):
    """A view for deleting events."""

    dialog_submit_actions = ('delete',)
    dialog_close_actions = ('cancel',)
    label = None
    data = None

    @Lazy
    def event(self):
        event_id = self.request.get('event_id', None)
        if event_id is None:
            return None
        try:
            return self.context.find(event_id)
        except KeyError:
            return None

    @Lazy
    def date(self):
        event_date = self.request.get('date', None)
        if event_date is None:
            return None
        return parse_date(event_date)

    def initDialog(self):
        super(DeleteEventDialog, self).initDialog()
        title = _(u'Delete event ${event}',
                  mapping={'event': self.event.title})
        self.ajax_settings['dialog']['title'] = translate(
            title, context=self.request)

    def nextURL(self):
        return (self.request.get('back_url') or
                absoluteURL(self.context, self.request))

    @Lazy
    def recurrent(self):
        return self.event.recurrence is not None

    @Lazy
    def day_title(self):
        preferences = ViewPreferences(self.request)
        dayformat = '%A, ' + preferences.dateformat
        return unicode(self.event.dtstart.strftime(dayformat))

    def modifyRecurrence(self, event, **kwargs):
        """Modify the recurrence rule of an event.

        If the event does not have any recurrences afterwards, it is removed
        from the parent calendar
        """
        rrule = event.recurrence
        new_rrule = rrule.replace(**kwargs)
        # This view requires the modifyEvent permission.
        event.recurrence = removeSecurityProxy(new_rrule)
        if not event.hasOccurrences():
            ICalendar(event).removeEvent(removeSecurityProxy(event))

    def deleteCurrent(self):
        exceptions = self.event.recurrence.exceptions + (self.date, )
        self.modifyRecurrence(self.event, exceptions=exceptions)

    def deleteFuture(self):
        self.modifyRecurrence(
            self.event, until=(self.date - datetime.timedelta(1)),
            count=None)

    def deleteAll(self):
        self.context.removeEvent(removeSecurityProxy(self.event))

    delete_handlers = {
        'all': deleteAll,
        'current': deleteCurrent,
        'future': deleteFuture,
        }

    def updateActions(self):
        super(DeleteEventDialog, self).updateActions()
        self.actions['delete'].addClass('button-ok')
        self.actions['cancel'].addClass('button-cancel')

    def update(self):
        self.data = {}
        if self.recurrent:
            self.fields = field.Fields(IDeleteRecurringEventForm)
            radio_widget = z3c.form.browser.radio.RadioFieldWidget
            self.fields['delete'].widgetFactory = radio_widget
        super(DeleteEventDialog, self).update()

    def getContent(self):
        return self.data

    @button.buttonAndHandler(_("Delete"), name='delete')
    def handleDelete(self, action):
        data, errors = self.extractData()
        if self.recurrent:
            handler = self.delete_handlers.get(data['delete'], None)
        else:
            handler = self.delete_handlers['all']

        if handler is not None:
            next_url = self.nextURL()
            handler(self)
            self.request.response.redirect(next_url.encode('utf-8'))
            self.ajax_settings['dialog'] = 'close'

    @button.buttonAndHandler(_("Cancel"))
    def handle_cancel_action(self, action):
        pass


class EventLinks(flourish.page.RefineLinksViewlet):
    """Manager for Action links in event views."""


class DeleteEventLinkViewlet(flourish.page.ModalFormLinkViewlet):

    @property
    def calendar(self):
        return ICalendar(self.event)

    @property
    def event(self):
        return self.context

    def nextURL(self):
        try:
            return self.view.nextURL()
        except AttributeError:
            return absoluteURL(self.calendar, self.request)

    @property
    def url(self):
        preferences = ViewPreferences(self.request)
        event = self.event
        start = event.dtstart.astimezone(preferences.timezone)
        url = '%s/delete.html?event_id=%s&date=%s&back_url=%s' % (
            absoluteURL(self.calendar, self.request),
            event.unique_id,
            start.strftime('%Y-%m-%d'),
            urllib.quote(self.nextURL()))
        return url




"""
Tests for generation scripts.
"""
from zope.interface import implements
from zope.keyreference.interfaces import IKeyReference
from zope.app.testing.setup import setUpAnnotations
from zope.app.publication.zopepublication import ZopePublication
from zope.site.folder import rootFolder
from zope.app.testing import setup


class ContextStub(object):
    """Stub for the context argument passed to evolve scripts.

        >>> from zope.app.generations.utility import getRootFolder
        >>> context = ContextStub()
        >>> getRootFolder(context) is context.root_folder
        True

    """

    class ConnectionStub(object):
        def __init__(self, root_folder):
            self.root_folder = root_folder
        def root(self):
            return {ZopePublication.root_name: self.root_folder}

    def __init__(self):
        self.root_folder = rootFolder()
        self.connection = self.ConnectionStub(self.root_folder)


_d = {}

class StupidKeyReference(object):
    implements(IKeyReference)
    key_type_id = 'StupidKeyReference'
    def __init__(self, ob):
        global _d
        self.id = id(ob)
        _d[self.id] = ob
    def __call__(self):
        return _d[self.id]
    def __hash__(self):
        return self.id
    def __cmp__(self, other):
        return cmp(hash(self), hash(other))


def catalogSetUp(test):
    """This code is deprecated, to be used only by old evolution tests."""
    setup.placefulSetUp()
    setup.setUpTraversal()
    setUpAnnotations()

    # this is code to set up the catalog for unit testing. it could
    # be extracted and put into general setup functionality

    # Make sure objects can be keyreferenced - necessary for int ids to
    # work:

    from zope.component import provideAdapter
    from persistent.interfaces import IPersistent
    provideAdapter(StupidKeyReference, [IPersistent], IKeyReference)

    # Provide the int id subscribers:

    from zope.component import provideHandler
    from zope.intid import addIntIdSubscriber, removeIntIdSubscriber
    from zope.location.interfaces import ILocation
    from zope.lifecycleevent.interfaces import IObjectAddedEvent
    from zope.lifecycleevent.interfaces import IObjectRemovedEvent
    provideHandler(addIntIdSubscriber,
                   [ILocation, IObjectAddedEvent])
    provideHandler(removeIntIdSubscriber,
                   [ILocation, IObjectRemovedEvent])

    # And the catalog subscribers:

    from zope.catalog import catalog
    from zope.catalog.interfaces import ICatalogIndex
    from zope.intid.interfaces import IIntIdAddedEvent,\
         IIntIdRemovedEvent
    from zope.lifecycleevent.interfaces import IObjectModifiedEvent
    provideHandler(catalog.indexAdded,
                   [ICatalogIndex, IObjectAddedEvent])
    provideHandler(catalog.indexDocSubscriber,
                   [IIntIdAddedEvent])
    provideHandler(catalog.reindexDocSubscriber,
                   [IObjectModifiedEvent])
    provideHandler(catalog.unindexDocSubscriber,
                   [IIntIdRemovedEvent])

def catalogTearDown(test):
    """This code is deprecated, to be used only by old evolution tests."""
    setup.placefulTearDown()
    _d.clear()
